import numpy as np
import test_001_to_from as ttf


def test_set_tube_nr():
    dino_gw = ttf.obscollection_dinozip_gw()
    dino_gw.gwobs.set_tube_nr(if_exists="replace")


def test_set_tube_nr_location():
    fews_gw_prod = ttf.obscollection_fews_lowmemory()
    fews_gw_prod.gwobs.set_tube_nr_location("location", if_exists="replace")


def test_get_modellayers_mf2005():
    import flopy

    modelname = "test_mf2005"
    ml = flopy.modflow.Modflow(modelname)
    # Model domain and grid definition
    Lx = 300000.0
    Ly = 400000.0
    ztop = 50.0
    zbot = -150.0
    nlay = 4
    nrow = 40
    ncol = 30
    delr = Lx / ncol
    delc = Ly / nrow
    botm = np.linspace(ztop, zbot, nlay + 1)
    # Create the discretization object
    flopy.modflow.ModflowDis(
        ml,
        nlay,
        nrow,
        ncol,
        delr=delr,
        delc=delc,
        top=ztop,
        botm=botm[1:],
        xul=0,
        yul=700000,
    )

    dino_gw = ttf.obscollection_dinozip_gw()
    dino_gw.gwobs.get_modellayers(ml)


def test_get_modellayers_mf6_structured():
    import flopy

    # Create the Flopy simulation object
    model_name = "test_mf6_structured"
    sim = flopy.mf6.MFSimulation(sim_name=model_name, exe_name="mf6", version="mf6")

    # Create the Flopy groundwater flow (gwf) model object
    model_nam_file = "{}.nam".format(model_name)
    gwf = flopy.mf6.ModflowGwf(sim, modelname=model_name, model_nam_file=model_nam_file)
    Lx = 300000.0
    Ly = 400000.0
    ztop = 50.0
    zbot = -150.0
    nlay = 4
    nrow = 40
    ncol = 30
    botm = np.linspace(ztop, zbot, nlay + 1)

    flopy.mf6.ModflowGwfdis(
        gwf,
        xorigin=0,
        yorigin=300000,
        nlay=4,
        nrow=nrow,
        ncol=ncol,
        delr=Lx / ncol,
        delc=Ly / nrow,
        top=ztop,
        botm=botm[1:],
    )
    dino_gw = ttf.obscollection_dinozip_gw()
    dino_gw.gwobs.get_modellayers(gwf)


def test_get_regis_layer():
    dino_gw = ttf.obscollection_dinozip_gw()
    dino_gw.gwobs.get_regis_layers()



# flake8: noqa
wsvv_pid = {
    "P.G.d9": "PrecipitationObs",  # Gemeten neerslag [mm] - dag (9-9 uur)
    "P.G.h": "PrecipitationObs",  # Gemeten neerslag [mm] - uur
    "P.TH.B": "PrecipitationObs",  # Doorval berekend [mm]
    "P.f.control": "PrecipitationObs",  # Neerslag verwachting [mm] (controle)
    "P.f.determ": "PrecipitationObs",  # Neerslag verwachting [mm] (deterministisch)
    "P.f.ensem": "PrecipitationObs",  # Neerslag verwachting [mm] (ensemble)
    "P.f.harmonie": "PrecipitationObs",  # Neerslag verwachting [mm] (harmonie)
    "P.f.hirlam": "PrecipitationObs",  # Neerslag verwachting [mm] (hirlam)
    "P.overschot": "PrecipitationObs",  # Neerslagoverschot [mm]
    "P.radar.5": "PrecipitationObs",  # Radarneerslag [mm] - 5min
    "P.radar.5.C": "PrecipitationObs",  # Radarneerslag [mm] - 5min gecorrigeerd
    "P.radar.5.O": "PrecipitationObs",  # Radarneerslag [mm] - 5min ongecorrigeerd
    "P.radar.d": "PrecipitationObs",  # Radarneerslag [mm] - dag
    "P.radar.d9": "PrecipitationObs",  # Radarneerslag [mm] - dag (9-9 uur)
    "P.radar.h": "PrecipitationObs",  # Radarneerslag [mm] - uur
    "Dsnow.f.harmonie": "PrecipitationObs",  # Sneeuwhoogte verwachting [m] (harmonie)
    "Pgrau.f.harmonie": "PrecipitationObs",  # Neerslag (hagel) verwachting [mm] harmonie
    "Psnow.f.harmonie": "PrecipitationObs",  # Neerslag (sneeuw) verwachting [mm] harmonie
    "Eact.d": "EvaporationObs",  # Actuele Verdamping [mm] - dag
    "Edef.d": "EvaporationObs",  # Verdampingstekort [mm] - dag
    "Eref.d": "EvaporationObs",  # Referentiegewasverdamping [mm] - dag
    "EF.GR.f.harmonie": "MeteoObs",  # Straling globaal verwacht [J/m2] (harmonie)
    "EF.LH.f.harmonie": "MeteoObs",  # Straling latente warmte verwacht [J/m2] (harmonie)
    "EF.LWR.f.harmonie": "MeteoObs",  # Straling lange golf verwacht [J/m2] (harmonie)
    "EF.SWR.f.harmonie": "MeteoObs",  # Straling korte golf verwacht [J/m2] (harmonie)
    "WGU.f.harmonie": "MeteoObs",  # Windstoot verwachting u-component [m/s] (harmonie)
    "WGV.f.harmonie": "MeteoObs",  # Windstoot verwachting v-component [m/s] (harmonie)
    "WR.G": "MeteoObs",  # Windrichting gemeten [graden]
    "WR.f.control": "MeteoObs",  # Windrichting verwacht [graden] (controle)
    "WR.f.determ": "MeteoObs",  # Windrichting verwacht [graden] (deterministisch)
    "WR.f.ensem": "MeteoObs",  # Windrichting verwacht [graden] (ensemble)
    "WS.G": "MeteoObs",  # Windsnelheid gemeten [m/s]
    "WS.f.control": "MeteoObs",  # Windsnelheid verwachting [m/s] (controle)
    "WS.f.determ": "MeteoObs",  # Windsnelheid verwachting [m/s] (deterministisch)
    "WS.f.ensem": "MeteoObs",  # Windsnelheid verwachting [m/s] (ensemble)
    "WSU10.f.harmonie": "MeteoObs",  # Windsnelheid verwachting u-component [m/s] harmonie
    "WSU10.f.hirlam": "MeteoObs",  # Windsnelheid verwachting u-component [m/s] (hirlam)
    "WSV10.f.harmonie": "MeteoObs",  # Windsnelheid verwachting v-component [m/s] harmonie
    "WSV10.f.hirlam": "MeteoObs",  # Windsnelheid verwachting v-component [m/s] (hirlam)
    "H.B": "WaterlvlObs",  # Waterhoogte berekend [mNAP]
    "H.B.d": "WaterlvlObs",  # Waterhoogte berekend [mNAP] - dag
    "H.B.h": "WaterlvlObs",  # Waterhoogte berekend [mNAP] - uur
    "H.G": "WaterlvlObs",  # Waterhoogte [mNAP]
    "H.G.AA.d": "WaterlvlObs",  # Waterhoogte alle [-] - dag
    "H.G.d": "WaterlvlObs",  # Waterhoogte [mNAP] - dag
    "H.G.h": "WaterlvlObs",  # Waterhoogte [mNAP] - uur
    "H.Max": "WaterlvlObs",  # Maximum waterpeil [mNAP]
    "H.S": "WaterlvlObs",  # Streefpeil [mNAP]
    "H.S.h": "WaterlvlObs",  # Streefpeil [mNAP] - uur
    "H.f": "WaterlvlObs",  # Waterhoogte verwachting [mNAP]
    "H.f.determ": "WaterlvlObs",  # Waterhoogte verwachting [mNAP] (ecmwf-deterministisch)
    "H.f.ecmwf": "WaterlvlObs",  # Waterhoogte verwachting [mNAP] (ecmwf-ensemble)
    "H.f.fr": "WaterlvlObs",  # Waterhoogte verwachting [mNAP] (fews rivieren)
    "H.f.matroos": "WaterlvlObs",  # Waterhoogte verwachting [mNAP] (matroos)
    "H.f.rws": "WaterlvlObs",  # Waterhoogte verwachting [mNAP] (rws)
    "H.f.wdij": "WaterlvlObs",  # Waterhoogte verwachting [mNAP] (wdij)
    "H.schuif.G": "WaterlvlObs",  # Schuifstand [mNAP]
    "H.wave": "WaterlvlObs",  # Golfhoogte [m]
    # "Hk": "Kruinhoogte [mNAP]",
    # "Hk.AA.d": "Kruinhoogte alle [-] - dag",
    # "Hk.d": "Kruinhoogte [mNAP] - dag",
    # "Hk.h": "Kruinhoogte [mNAP] - uur",
    # "Hopv.B": "Opvoerhoogte berekend [mNAP]",
    # "Hpbe.1b": "Peilevaluatie peil eerste bovengrens [mNAP]",
    # "Hpbe.1o": "Peilevaluatie peil eerste ondergrens [mNAP]",
    # "Hpbe.2b": "Peilevaluatie peil tweede bovengrens [mNAP]",
    # "Hpbe.2o": "Peilevaluatie peil tweede ondergrens [mNAP]",
    # "Hu.G": "Uitwateringspeil [mNAP]",
    "Hwk": "WaterlvlObs",  # Hoogte waterkolom [m]
    "Hwk.H1": "WaterlvlObs",  # Overstortende straal H1 [m]
    "Hwk.H2": "WaterlvlObs",  # Hoogteverschil na stuw H2 [m]
    "Hwk.PA.G.Barodiver": "WaterlvlObs",  # Hoogteverschil na stuw H2 [m]
    "Hwk.PW.G": "WaterlvlObs",  # Gemeten waterdruk [m]
    "Hwk.V.PB": "WaterlvlObs",  # Waterstandsafwijking PB [m]
    "Hwk.V.SP": "WaterlvlObs",  # Waterstandsafwijking SP [m]
    # "Q.B": "WaterlvlObs",  # Debiet FEWS [m3/s]
    # "Q.B.d": "WaterlvlObs",  # Debiet FEWS [m3/s] - dag
    # "Q.B.h": "WaterlvlObs",  # Debiet FEWS [m3/s] - uur
    # "Q.G": "WaterlvlObs",  # Debiet [m3/s]
    # "Q.G.d": "WaterlvlObs", # Debiet [m3/s] - dag
    # "Q.G.h": "WaterlvlObs",  # Debiet [m3/s] - uur
    # "Q.f": "WaterlvlObs",  # Debiet verwachting [m3/s]
    # "Q.f.determ": "WaterlvlObs",  # Debiet verwachting [m3/s] (ecmwf-deterministisch)
    # "Q.f.ecmwf": "WaterlvlObs",  # Debiet verwachting [m3/s] (ecmwf-ensemble)
    # "Q.f.fr": "WaterlvlObs",  # Debiet verwachting [m3/s] (fews rivieren)
    # "Q.f.rws": "WaterlvlObs",  # Debiet verwachting [m3/s] (rws)
    # "Q.f.totaal": "WaterlvlObs",  # Debiet verwachting [m3/s] (totaal)
    # "Qafval.G": "WaterlvlObs",  # Debiet afvalwater [m3/h]
    # "Qafval.G.d": "WaterlvlObs",  # Debiet afvalwater [m3/h] - dag
    # "Qafval.G.h": "WaterlvlObs",  # Debiet afvalwater [m3/h] - uur
    "T.ow.G": "WaterQualityObs",  # Temperatuur oppervlaktewater [C]
    "O2.gehalte": "WaterQualityObs",  # Zuurstof gehalte [mg/l]
    "O2.saturatie": "WaterQualityObs",  # Zuurstof saturatie [%]
    "PA.G": "MeteoObs",  # Luchtdruk gemeten [hPa]
    "PA.G.bar": "MeteoObs",  # Luchtdruk gemeten [bar]
    "PA.f.harmonie": "MeteoObs",  # Luchtdruk verwachting [hPa] (harmonie)
    "PKEL.V": "MeteoObs",  # Drukverschil Keller (P1-P2)
    "RH.G": "MeteoObs",  # Luchtvochtigheid [%]
    "RH.f.harmonie": "MeteoObs",  # Relatieve luchtvochtigheid [%] (harmonie)
    "T.air.G": "MeteoObs",  # Temperatuur lucht [C]
    "T.air.f.control": "MeteoObs",  # Temperatuur verwacht (controle)
    "T.air.f.determ": "MeteoObs",  # Temperatuur verwacht (deterministisch)
    "T.air.f.ensem": "MeteoObs",  # Temperatuur verwacht (ensemble)
    "T.air.f.harmonie": "MeteoObs",  # Temperatuur verwacht (harmonie)
    # "PW.G": "Waterspanning [hPa]",
    # "PZ.G": "Zuigspanning [kPa]",
    # "SM.G": "Bodemvochtigheid [m3/m3]",
    # "T.soil.G": "Temperatuur bodem [C]",
    "SS.G": "WaterlvlObs",  # Stroomsnelheid gemeten [m/s]
    "SS.G.d": "WaterlvlObs",  # Stroomsnelheid gemeten [m/s] - dag
    "SS.G.h": "WaterlvlObs",  # Stroomsnelheid gemeten [m/s] - uur
    "GW.G": "GroundwaterObs",  # Grondwaterstand [mNAP]
    "GW.G.hm": "WaterlvlObs",  # Grondwaterstand - handmeting [mNAP]
    "SH": "GroundwaterObs",  # Stijghoogte [mNAP]
    "T.gw.G": "WaterQualityObs",  # Temperatuur grondwater [C]
}

hhnk_pid = {
    "windrichting.meting": "MeteoObs",  # windrichting gemeten [graden]
    "windrichting.rws": "MeteoObs",  # windrichting voorspeld rws [graden]
    "windrichting.voorsp.det": "MeteoObs",  # windrichting voorspeld (det) [graden]
    "windrichting.voorsp.ens": "MeteoObs",  # windrichting voorspeld (ens) [graden]
    "Rns.voorsp": "MeteoObs",  # netto kortgolvige straling (Joules/m2) [J/m2]
    "LH.voorsp": "MeteoObs",  # Latente warmte (Joules/m2) [J/m2]
    "windsnelheid.meting": "MeteoObs",  # windsnelheid gemeten [m/s]
    "Wind.voorsp.u": "MeteoObs",  # voorspelde windsnelheid (u-horiz.) [m/s]
    "Wind.voorsp.v": "MeteoObs",  # voorspelde windsnelheid (v-vert.) [m/s]
    "wind.meting.max": "MeteoObs",  # windsnelheid gemeten max [m/s]
    "windsnelheid.rws": "MeteoObs",  # windsnelheid voorspeld rws [m/s]
    "windsnelheid.voorsp.det": "MeteoObs",  # windsnelheid voorspeld (det) [m/s]
    "windsnelheid.voorsp.ens": "MeteoObs",  # windsnelheid voorspeld (ens) [m/s]
    "WNS9026": "MeteoObs",  # luchtdruk [hPa]",
    "GELDHD.meting": "WaterQualityObs",  # geleidbaarheid (EGV) [mS/m]
    "GELDHD.meting.hand": "WaterQualityObs",  # geleidbaarheid (EGV) handmeting [mS/m]
    "EGV.meting": "WaterQualityObs",  # geleidbaarheid (EGV) [mS/cm]
    "EGVms_cm.meting": "WaterQualityObs",  # geleidbaarheid (EGV mS/cm) [mS/cm]
    "EGVms.meting": "WaterQualityObs",  # geleidbaarheid (EGV mS) [mS]
    "EGVms_m.meting": "WaterQualityObs",  # geleidbaarheid (EGV mS/m) [mS/m]
    "EGVms_m.meting.hand": "WaterQualityObs",  # geleidbaarheid handmeting (EGV mS/m) [mS/m]
    "O2.geh.meting": "WaterQualityObs",  # zuurstofgehalte [mg/l]
    "O2.sat.meting": "WaterQualityObs",  # zuurstofsaturatie [%]
    "CL.berekend": "WaterQualityObs",  # chloridegehalte berekend [mg/l]
    "T.water.meting": "WaterQualityObs",  # watertemperatuur [oC]
    "T.water.meting.5min": "WaterQualityObs",  # watertemperatuur [oC]
    "T.water.meting.hand": "WaterQualityObs",  # watertemperatuur handmeting [oC]
    "GW.meting": "GroundWaterObs",  # grondwaterhoogte gemeten [m]
    "H.meting.ref": "GroundwaterObs",  # cm tov referentiehoogte [cm]
    "H.meting.5min": "GroundWaterObs",  # grondwaterhoogte gemeten [m]
    "GW.meting.1%": "GroundWaterObs",  # 1% percentiel grondwater [m]
    "GW.meting.10%": "GroundWaterObs",  # 10% percentiel grondwater [m]
    "GW.meting.90%": "GroundWaterObs",  # 90% percentiel grondwater [m]
    "GW.meting.25%": "GroundWaterObs",  # 25% percentiel grondwater [m]
    "GW.meting.75%": "GroundWaterObs",  # 75% percentiel grondwater [m]
    "GW.meting.99%": "GroundWaterObs",  # 99% percentiel grondwater [m]
    "H.berekend": "WaterlvlObs",  # waterhoogte berekend [m]
    "H.streef": "WaterlvlObs",  # streefpeil [m]
    "H.streef.boven": "WaterlvlObs",  # bovengrens streefpeil [m]
    "H.streef.onder": "WaterlvlObs",  # ondergrens streefpeil [m]
    "H.meting": "WaterlvlObs",  # waterhoogte [m]
    "H.meting.hand": "WaterlvlObs",  # waterhoogte handmeting [m]
    "H.meting.org": "WaterlvlObs",  # waterhoogte origineel [m]
    "H.meting.int": "WaterlvlObs",  # waterhoogte geinterpoleerd [m]
    "H.meting.gebied": "WaterlvlObs",  # gebiedsgemiddelde waterhoogte [m]
    "H.meting.1%": "WaterlvlObs",  # 1% percentiel waterhoogte [m]
    "H.meting.10%": "WaterlvlObs",  # 10% percentiel waterhoogte [m]
    "H.meting.90%": "WaterlvlObs",  # 90% percentiel waterhoogte [m]
    "H.meting.25%": "WaterlvlObs",  # 25% percentiel waterhoogte [m]
    "H.meting.75%": "WaterlvlObs",  # 75% percentiel waterhoogte [m]
    "H.meting.99%": "WaterlvlObs",  # 99% percentiel waterhoogte [m]
    "H.meting.dag": "WaterlvlObs",  # waterhoogte dagwaarden [m]
    "H.astronomisch": "WaterlvlObs",  # getij astronomisch [m]
    "H.getij.meting": "WaterlvlObs",  # getij gemeten [m]
    "H.getij.meting.opzet": "WaterlvlObs",  # getij opzet gemeten [m]
    "H.getij.hmcn": "WaterlvlObs",  # getij voorspeld hmcn [m]
    "H.getij.rws": "WaterlvlObs",  # getij voorspeld rws [m]
    "H.getij.wdij": "WaterlvlObs",  # getij voorspeld wdij [m]
    "H.getij.opzet.rws": "WaterlvlObs",  # getij opzet voorspeld rws [m]
    "WNS9688": "WaterlvlObs",  # waterhoogte [m]
    # "inslagpeil": "2e inslagpeil [m]
    # "Bov.buis": "bovenkant buis [m]",
    # "Maaiveld": "maaiveld [m]",
    "P.meting.1m": "PrecipitationObs",  # neerslag gemeten (1 min) [mm]
    "P.meting.5m": "PrecipitationObs",  # neerslag gemeten (5 min) [mm]
    "P.meting.10m": "PrecipitationObs",  # neerslag gemeten (10 min) [mm]
    "P.meting.1h": "PrecipitationObs",  # neerslag gemeten (1 uur) [mm]
    "P.meting.24h": "PrecipitationObs",  # neerslag gemeten (24 uur) [mm]
    "P.radar.2m": "PrecipitationObs",  # neerslagradar (2 min) [mm/5min]
    "P.radar.5m": "PrecipitationObs",  # neerslagradar (5 min) [mm/5min]
    "P.radar.hist": "PrecipitationObs",  # neerslag (ongecalibreerde radar) [mm/hr]
    "P.radar.cal": "PrecipitationObs",  # neerslag (gecalibreerde radar) [mm/hr]
    "P.voorsp": "PrecipitationObs",  # neerslag voorspeld [mm/hr]
    "P.voorsp.ens": "PrecipitationObs",  # neerslag voorspeld (ens) [mm/hr]
    "P.voorsp.ens.min": "PrecipitationObs",  # neerslag voorspeld (ens) min [mm/hr]
    "P.voorsp.ens.max": "PrecipitationObs",  # neerslag voorspeld (ens) max [mm/hr]
    "P.voorsp.ens.median": "PrecipitationObs",  # neerslag voorspeld (ens) mediaan [mm/hr]
    "P.voorsp.ens.10": "PrecipitationObs",  # neerslag voorspeld (ens) 10 [mm/hr]
    "P.voorsp.ens.90": "PrecipitationObs",  # neerslag voorspeld (ens) 90 [mm/hr]
    "P.voorsp.ctr": "PrecipitationObs",  # neerslag voorspeld (ctr) [mm/hr]
    "P.voorsp.det": "PrecipitationObs",  # neerslag voorspeld (det) [mm/hr]
    "P.radar.1h": "PrecipitationObs",  # neerslagradar (1 uur) [mm/hr]
    "P.radar.24h": "PrecipitationObs",  # neerslagradar (24 uur) [mm/24hr]
    "E.meting": "EvaporationObs",  # verdamping berekend KNMI (Makkink) [mm]
    "E.ref.Makkink": "EvaporationObs",  # verdamping voorspeld (Makkink gewas ref) [mm]
    "Neerslagoverschot": "MeteoObs",  # Berekende neerslagoverschot (mm/dag) [mm]
    "Neerslagoverschot_cumulatief": "MeteoObs",  # Berekende cumulatieve neerslagoverschot (mm) [mm]
    "T.lucht.meting": "MeteoObs",  # luchttemperatuur [oC]
    "T.meting": "MeteoObs",  # luchttemperatuur [oC]
    "T.voorsp": "MeteoObs",  # temperatuur verwacht [oC]
    "T.voorsp.ens": "MeteoObs",  # temperatuur voorspeld (ens) [oC]
    "T.voorsp.det": "MeteoObs",  # temperatuur voorspeld (det) [oC]
    # "T.golf.meting": "golfperiode [-]",
    # "Hz.meting": "frequentie [Hz]",
    # "H.golf.meting": "golfhoogte [m]",
    # "Stuw.klep.meting": "klephoogte [m]",
    # "H.meting.proc": "vullingsgraad [%]",
    # "H.meting.percentage": "niveau percentage [%]",
    # "Druk.meting": "druk gemeten [kPa]",
    # "P.druk.meting": "druk [mBar]",
    # "WNS2368": "debiet [m3/h]",
    # "WNS2369.sim": "debiet voorspeld [m3/h]",
    # "WNS2369.h": "verpompt volume uur [m3/h]",
    # "WNS2369.h.pred": "verpompt volume uur voorspeld [m3/h]",
    # "WNS2369.h.upp": "verpompt volume uur voorspeld bovengrens [m3/h]",
    # "WNS2369.h.low": "verpompt volume uur voorspeld ondergrens [m3/h]",
    # "Q.meting": "debiet gemeten [m3/s]",
    # "Q.berekend": "debiet berekend [m3/s]",
    # "Q.berekend.totNZK": "debiet totaal NZK [m3/s]",
    # "Q.berekend.totNZK.1h": "debiet totaal NZK (1 uur) [m3/s]",
    # "Q.advies": "debiet advies [m3/s]",
    # "WNS2369.2min": "verpompt volume 2min [m3/2min]",
    # "WNS2369.1min": "verpompt volume min [m3/min]",
    # "Q.berekend.dag": "debiet berekend dag [m3/dag]",
    # "WNS2369.d": "verpompt volume dag [m3/dag]",
    # "AfwijkingInslagpeil": "afwijking peil [min]",
    # "AfwijkingNorm_kwartaal": "tijd niet voldaan aan afnameverplichting [min]",
    # "AfwijkingNorm": "voldoet aan afnameverplichting [min]",
    # "AfwijkingInslagpeil_kwartaal": "tijd boven 2e inslagpeil [min]",
    # "AfwijkingRVW_dag": "Percentage RVW >50% over 24 uur [%]",
    # "AfwijkingRVW_dag_nat": "Percentage RVW >50% over 24 uur nat [%]",
    # "AfwijkingRVW_dag_droog": "Percentage RVW >50% over 24 uur droog [%]",
}

wli_pid = {
    "WATDTE": "GroundwaterObs",  # Waterdiepte [m]
    "DRUK": "GroundwaterObs",  # Drukhoogte [m]
    "STIJGHTE": "GroundwaterObs",  # Stijghoogte [m]
    "STIJGHTE_sim": "GroundwaterObs",  # Stijghoogte gesimuleerd [m]
    "STIJGHTE_res": "GroundwaterObs",  # Resultaat toetsing PASTAS
    "WATHTE": "WaterlvlObs",  # Waterhoogte [m]
    # "HEFHTE": "Hefhoogte [m]",
    # "KRUINHTE": "Kruinhoogte [m]",
    # "STREEFHTE": "Streefhoogte [m]",
    # "MAAIVELDHTE": "Maaiveldhoogte [m]",
    # "KRUINREL": "Kruinoogte [%]"
    # "STROOMSHD": "Stroomsnelheid [m/s]",
    # "Q": "Afvoer [m3/s]",
    # "Q_res" : "Resultaat toetsing PASTAS Afvoer res"
    # "STREEFQ": "Afvoer streefwaarde [m3/s]",
    # "Q_sim": "Afvoer gesimuleerd [m3/s]",
    # "Q_sim.fc": "Afvoer voorspeld is gemiddelde van voorspellingen (sim.fc) [m3/s]",
    # "Q_sim_q05.fc": "Afvoer gesimuleerd 5%-waarde of ondergrens 90% betrouwbaarheidsinterval [m3/s]",
    # "Q_sim_q25.fc": "Afvoer gesimuleerd 25%-waarde of ondergrens 50% betrouwbaarheidsinterval [m3/s]",
    # "Q_sim_q75.fc": "Afvoer gesimuleerd 75%-waarde of bovengrens 50% betrouwbaarheidsinterval [m3/s]",
    # "Q_sim_q95.fc": "Afvoer gesimuleerd 95%-waarde of bovengrens 90% betrouwbaarheidsinterval [m3/s]",
    "GELDHD": "WaterQualityObs",  # Geleidbaarheid (MON) [uS/cm]
    "O2SAT": "WaterQualityObs",  #  Zuurstof saturatie [%]
    "O2": "WaterQualityObs",  # Zuurstofgehalte [mg/l]
    "NEERSG": "PrecipitationObs",  # Neerslag [mm]
    "NEERSG.tekort": "MeteoObs",  # Neerslagtekort [mm]
    "NEERSG.voorsp.det": "PrecipitationObs",  # Neerslag voorspeld deterministisch [mm]
    "NEERSG.voorsp.ens": "PrecipitationObs",  # Neerslag voorspeld ensemble [mm]
    "VERDPG": "EvaporationObs",  # Verdamping [mm]
    "LUCHTDK": "MeteoObs",  # Luchtdruk [m]
    "WINDRTG": "MeteoObs",  # Wind richting [degrees]
    "WINDSHD": "MeteoObs",  # Windhsnelheid [m/s]
    "GLOBSLG": "MeteoObs",  # Globale straling [Joules/cm2]
    "T": "MeteoObs",  # Temperatuur [C]
    "T.voorsp.det": "MeteoObs",  # Temperatuur voorspeld deterministisch [C]
    "T.voorsp.ens": "MeteoObs",  # Temperatuur voorspeld ensemble [C]
    "RELTVLVTHD": "MeteoObs",  # Relatieve luchtvochtigheid [%]
}

pid = {"wsvv": wsvv_pid, "hhnk": hhnk_pid, "wli": wli_pid}



import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.gridspec import GridSpec
from tqdm.auto import tqdm

from ..observation import GroundwaterObs
from . import accessor

logger = logging.getLogger(__name__)


@accessor.register_obscollection_accessor("plots")
class CollectionPlots:
    def __init__(self, oc_obj):
        """Object containing plotting methods for ObsCollections.

        Parameters
        ----------
        oc : ObsCollection
            ObsCollection instance
        """
        self._obj = oc_obj

    def interactive_plots(
        self,
        savedir="figures",
        tmin=None,
        tmax=None,
        per_location=True,
        **kwargs,
    ):
        """Create interactive plots of the observations using bokeh.

        Parameters
        ----------
        savedir : str
            directory used for the folium map and bokeh plots
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        per_location : bool, optional
            if True plot multiple observations at the same location in one
            figure
        **kwargs :
            will be passed to the Obs.interactive_plot method, options
            include:

            - cols : list of str or None
            - hoover_names : list of str
            - plot_freq : list of str
            - plot_legend_names : list of str
            - markers : list of str
            - hoover_names : list of str
            - plot_colors : list of str
            - ylabel : str
            - add_screen_to_legend : boolean
        """
        _color_cycle = (
            "blue",
            "olive",
            "lime",
            "red",
            "orange",
            "yellow",
            "purple",
            "silver",
            "powderblue",
            "salmon",
            "tan",
        )

        if per_location:
            plot_names = self._obj.groupby("location").count().index
        else:
            plot_names = self._obj.index

        for name in plot_names:
            if per_location:
                oc = self._obj.loc[self._obj.location == name, "obs"].sort_index()
            else:
                oc = self._obj.loc[[name], "obs"]

            p = None
            for i, o in enumerate(oc.values):
                if i == 10:
                    raise NotImplementedError(
                        "cannot add more than 10 lines to a single plot"
                    )
                try:
                    p = o.plots.interactive_plot(
                        savedir=savedir,
                        p=p,
                        tmin=tmin,
                        tmax=tmax,
                        plot_colors=[_color_cycle[i]],
                        return_filename=False,
                        **kwargs,
                    )
                    logger.debug(f"created iplot -> {o.name}")
                except ValueError:
                    logger.error(f"{o.name} has no data between {tmin} and {tmax}")
                    o.meta["iplot_fname"] = None

    def interactive_map(
        self,
        plot_dir="figures",
        m=None,
        tiles="OpenStreetMap",
        fname=None,
        per_location=True,
        color="blue",
        legend_name=None,
        add_legend=True,
        map_label="",
        map_label_size=20,
        col_name_lat="lat",
        col_name_lon="lon",
        zoom_start=13,
        popup_width=620,
        create_interactive_plots=True,
        **kwargs,
    ):
        """Create an interactive map with interactive plots using folium and
        bokeh.

        Notes
        -----
        Some notes on this method:

        - if you want to have multiple obs collections on one folium map, only
          the last one should have add_legend = True to create a correct legend
        - the color of the observation point on the map is now the same color
          as the line of the observation measurements. Also a built-in color
          cycle is used for different measurements at the same location.

        Parameters
        ----------
        plot_dir : str
            directory used for the folium map and bokeh plots
        m : folium.Map, str, optional
            current map to add observations too, if None a new map is created
        tiles : str, optional
            background tiles, default is openstreetmap
        fname : str, optional
            name of the folium map
        per_location : bool, optional
            if True plot multiple observations at the same location in one
            figure
        color : str, optional
            color of the observation points on the map
        legend_name : str, optional
            the name of the observation points shown in the map legend
        add_legend : boolean, optional
            add a legend to a plot
        map_label : str, optional
            add a label to the observations on the map, the label should be
            1. the attribute of an observation
            2. the key in the meta attribute of the observation
            3. a generic label for each observation in this collection.
            A label is only added if map_label is not ''. The default is ''.
        map_label_size : int, optional
            label size of the map_label in pt.
        col_name_lat : str, optional
            name of the column in the obs_collection dic with the lat values
            of the observation points
        col_name_lon : str, optional
            see col_name_lat
        zoom_start : int, optional
            start zoom level of the folium ma
        create_interactive_plots : boolean, optional
            if True interactive plots will be created, if False the iplot_fname
            in the meta ditctionary of the observations is used.
        popup_width : int, optional
            popup width in pixels, optional.
        **kwargs :
            will be passed to the interactive_plots method options are:

            - cols : list of str or None
            - hoover_names : list of str
            - plot_legend_names : list of str
            - plot_freq : list of str
            - markers : list of str
            - hoover_names : list of str
            - plot_colors : list of str
            - ylabel : str
            - add_screen_to_legend : boolean
            - tmin : dt.datetime
            - tmax : dt.datetime

        Returns
        -------
        m : folium.Map
            the folium map
        """

        import branca
        import folium
        from folium.features import DivIcon

        # check for empty observations
        if all([o.empty for o in self._obj.obs.values]):
            logger.warning("all observations in the collection are empty")
            for oname in self._obj.index:
                self._obj._set_metadata_value(
                    oname, "iplot_fname", None, add_to_meta=True
                )
            empty_obs = True
        else:
            empty_obs = False

        # create interactive bokeh plots
        if create_interactive_plots and not empty_obs:
            self._obj.plots.interactive_plots(
                savedir=plot_dir, per_location=per_location, width=popup_width, **kwargs
            )

        # check if observation collection has lat and lon values
        if (col_name_lat not in self._obj.columns) and (
            col_name_lon not in self._obj.columns
        ):
            self._obj.geo.set_lat_lon()

        # determine start location of map
        northing = np.mean(
            (self._obj[col_name_lat].min(), self._obj[col_name_lat].max())
        )
        easting = np.mean(
            (self._obj[col_name_lon].min(), self._obj[col_name_lon].max())
        )

        # create map if no map is given
        if m is None:
            m = folium.Map([northing, easting], zoom_start=zoom_start, tiles=tiles)

        # get oc name if no legend name is given
        if legend_name is None:
            legend_name = self._obj.name

        # add the point observations with plots to the map
        group_name = '<span style=\\"color: {};\\">{}</span>'.format(color, legend_name)
        group = folium.FeatureGroup(name=group_name)

        if per_location:
            plot_names = self._obj.groupby("location").count().index
        else:
            plot_names = self._obj.index

        for name in plot_names:
            if per_location:
                oc = self._obj.loc[self._obj.location == name, "obs"].sort_index()
                o = oc.iloc[-1]
                name = o.name
            else:
                o = self._obj.loc[name, "obs"]

            if o.meta["iplot_fname"] is not None:
                with open(o.meta["iplot_fname"], "r") as f:
                    bokeh_html = f.read()

                iframe = branca.element.IFrame(
                    html=bokeh_html, width=popup_width, height=int(popup_width / 1.5)
                )
                popup = folium.Popup(iframe, max_width=popup_width)

                folium.CircleMarker(
                    [
                        self._obj.loc[o.name, col_name_lat],
                        self._obj.loc[o.name, col_name_lon],
                    ],
                    icon=folium.Icon(icon="signal"),
                    fill=True,
                    color=color,
                    popup=popup,
                ).add_to(group)

                if map_label != "":
                    if map_label in o._get_meta_attr():
                        map_label_val = getattr(o, map_label)
                    elif map_label in o.meta.keys():
                        map_label_val = o.meta[map_label]
                    else:
                        map_label_val = map_label
                    folium.map.Marker(
                        [
                            self._obj.loc[name, col_name_lat],
                            self._obj.loc[name, col_name_lon],
                        ],
                        icon=DivIcon(
                            icon_size=(150, 36),
                            icon_anchor=(0, 0),
                            html='<div style="font-size: %ipt">%s</div>'
                            % (map_label_size, map_label_val),
                        ),
                    ).add_to(group)
            else:
                logger.info(f"no iplot available for {o.name}")
                folium.CircleMarker(
                    [
                        self._obj.loc[o.name, col_name_lat],
                        self._obj.loc[o.name, col_name_lon],
                    ],
                    icon=folium.Icon(icon="signal"),
                    fill=True,
                    color=color,
                ).add_to(group)

        group.add_to(m)

        # add legend
        if add_legend:
            folium.map.LayerControl("topright", collapsed=False).add_to(m)

        # save map
        # filename and path
        if fname is not None:
            if not fname.endswith(".html"):
                fname = fname + ".html"
            if not os.path.exists(plot_dir):
                os.mkdir(plot_dir)
            m.save(os.path.join(plot_dir, fname))

        return m

    def section_plot(
        self,
        tmin=None,
        tmax=None,
        cols=(None,),
        section_colname_x=None,
        section_label_x=None,
        section_well_layout_color="gray",
        section_markersize=100,
        ylabel="auto",
        fn_save=None,
        check_obs_close_to_screen_bottom=True,
        plot_well_layout_markers=True,
        plot_obs=True,
    ):
        """Create plot with well layout (left) en observations (right).

        Parameters
        ----------
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        cols : tuple of str or None, optional
                the columns of the observation to plot. The first numeric column
                is used if cols is None, by default None.
        section_colname_x : str, optional
            column used for x position on section plot, when None order
            collection is used
        section_label_x : str, optional
            label applied to x-axis in section plot
        section_well_layout_color : str, optional
            color of well layout, default is gray
        section_markersize : int, optional
            size of makers in sectionplot
        ylabel: str or list, optional
            when 'auto' column unit in collection is ylabel, otherwise first
            element of list is label of section plot, second element of observation plot
        fn_save : str, optional
            filename to save plot
        check_obs_close_to_screen_bottom : bool, optional
            plots a horizontal line when minimum observation is close to screen_bottom
        plot_well_layout_markers : bool, optional
            plots ground level, top tube, screen levels and sandtrap via
            makers. Default is True
        plot_obs : bool, optional
            Plots observation. Default is True

        TODO:
            - speficy colors via extra column in ObsCollection
            - addtional visual checks:
                maximum observation is close to or above ground level,
                maximum observation is close to or above tube top
                minimum observation is close to or below tube bottom (sand trap)
            - include some interactive Bokeh fancy
            - apply an offset when two wells are at same location
            - limit y-axis of section plot to observations only
            - remove the checking (if obs are near bottom) from this function
            - moving the legend outside the plot
            - set xlim of observation plot more tight when tmin is not specified
        """

        # prepare column for x location in section plot
        if section_colname_x is None:
            # use order in ObsCollection
            plot_x = np.arange(len(self._obj))
        else:
            plot_x = self._obj[section_colname_x].to_numpy()

        # create figure
        fig = plt.figure(figsize=(15, 5))
        if plot_obs:
            # make figure with section plot on left, and righthand plot for obs
            gs = GridSpec(
                1,
                2,
                width_ratios=[1, 3],
            )
            ax_section = fig.add_subplot(gs[0])
            ax_obs = fig.add_subplot(gs[1])
            axes = [ax_section, ax_obs]
        else:
            # make figure with only one plot on left, observations are not plotted
            ax_section = fig.add_subplot(111)
            axes = [ax_section]

        if plot_well_layout_markers:
            # plot well layout via markers
            ax_section.scatter(
                plot_x,
                self._obj.tube_top.values,
                section_markersize,
                label="tube top",
                marker="*",
                facecolors="none",
                color=section_well_layout_color,
            )
            ax_section.scatter(
                plot_x,
                self._obj.ground_level.values,
                section_markersize,
                label="ground level",
                marker="_",
                color=section_well_layout_color,
            )
            ax_section.scatter(
                plot_x,
                self._obj.screen_top.values,
                section_markersize / 2,
                label="screen top",
                marker="x",
                color=section_well_layout_color,
            )
            ax_section.scatter(
                plot_x,
                self._obj.screen_bottom.values,
                section_markersize,
                label="screen bottom",
                marker="+",
                color=section_well_layout_color,
            )

        # loop over all wells, plot observations and details in section plot
        for counter, name in enumerate(self._obj.index):
            if plot_obs:
                # which column to plot?
                cols = list(cols)
                for i, col in enumerate(cols):
                    if col is None:
                        cols[i] = self._obj.loc[
                            name, "obs"
                        ]._get_first_numeric_col_name()

                # create plot dataframe
                plot_df = self._obj.loc[name, "obs"][tmin:tmax][cols].copy()
                if plot_df.empty or plot_df[cols].isna().all().all():
                    logger.warning(f"{name} has no data between {tmin} and {tmax}")
                    continue

                # PART 1: plot timeseries of observations
                # one or multiple columns to plot?
                if len(cols) == 1:
                    p = ax_obs.plot(plot_df, label=name)
                else:
                    for col in cols:
                        p = ax_obs.plot(plot_df[col], label=f"{name}, {col}")
                plot_color = p[0].get_color()

                if check_obs_close_to_screen_bottom:
                    # add horizonal line to plot when minimum observation in first plot
                    # column is close to bottom of screen
                    offset = 0.1
                    if self._obj.loc[name, "screen_bottom"] > (
                        plot_df[cols[0]].dropna().min() - offset
                    ):
                        ax_obs.axhline(
                            y=self._obj.loc[name, "screen_bottom"],
                            ls="--",
                            lw=4,
                            alpha=0.5,
                            label=(
                                f"screen bottom of {name}\n"
                                "is close to minimum observation"
                            ),
                            color=plot_color,
                        )
            else:
                plot_color = "hotpink"

            # PART 2: fancy section plot with lines along tube

            # highlight filter on section plot
            ax_section.plot(
                [plot_x[counter]] * 2,
                [
                    self._obj.loc[name, "screen_top"],
                    self._obj.loc[name, "screen_bottom"],
                ],
                color="k",
                lw=3,
                ls="-",
            )

            # highlight blind tube on section plot
            ax_section.plot(
                [plot_x[counter]] * 2,
                [self._obj.loc[name, "screen_top"], self._obj.loc[name, "tube_top"]],
                color=plot_color,
                lw=3,
            )

            # add sandtrap when present
            if "tube_bottom" in self._obj.columns:
                ax_section.plot(
                    [plot_x[counter]] * 2,
                    [
                        self._obj.loc[name, "screen_bottom"],
                        self._obj.loc[name, "tube_bottom"],
                    ],
                    color=plot_color,
                    lw=3,
                )

            if plot_obs:
                # PART 3: fancy section plot with bandwith of observations
                ax_section.scatter(
                    plot_x[counter],
                    plot_df.quantile(q=0.95),
                    section_markersize,
                    marker=(3, 0, 0),
                    color="blue",
                    alpha=0.5,
                    label="95% observation",
                    zorder=100,
                )
                ax_section.scatter(
                    plot_x[counter],
                    plot_df.median(),
                    section_markersize,
                    marker=(4, 0, 90),
                    color="green",
                    alpha=0.5,
                    label="median",
                    zorder=100,
                )
                ax_section.scatter(
                    plot_x[counter],
                    plot_df.quantile(q=0.05),
                    section_markersize,
                    marker=(3, 0, 180),
                    color="red",
                    alpha=0.5,
                    label="5% observation",
                    zorder=100,
                )

            logger.info(f"created sectionplot -> {name}")

        # layout
        if section_label_x is None:
            # use name as xtick and rotate
            ax_section.set_xticks(
                plot_x,
                self._obj.index,
                rotation="vertical",
                fontsize="small",
            )
        else:
            ax_section.set_xlabel(section_label_x)

        if plot_obs:
            ax_obs.set_xlim(left=tmin, right=tmax)

            # rotate labels on observation axis
            ax_obs.set_xticks(
                ax_obs.get_xticks(),
                ax_obs.get_xticklabels(),
                rotation="vertical",
                fontsize="small",
            )

        if ylabel == "auto":
            # has collection uniform unit?
            if len(self._obj.unit.unique()) == 1:
                ylabel = self._obj.unit.unique()[0]
            else:
                logger.error(
                    "Collection has more than one unit. Plot has both on one y-axis."
                )
                ylabel = ", ".join(map(str, self._obj.unit.unique()))
            for ax in axes:
                ax.set_ylabel(ylabel)
        else:
            try:
                ax_section.set_ylabel(ylabel[0])
                ax_obs.set_ylabel(ylabel[1])
            except ValueError:
                logger.error(f"Invalid value for ylabel {ylabel}. Plot has no ylabels.")

        # add layout to both plots
        for ax in axes:
            ax.grid()

            # simple legend
            handles, labels = ax.get_legend_handles_labels()
            by_label = dict(zip(labels, handles))
            ax.legend(by_label.values(), by_label.keys())

        fig.suptitle(self._obj.name)

        if fn_save is not None:
            try:
                fig.tight_layout()
                fig.savefig(
                    fn_save,
                    dpi=250,
                )
            except ValueError:
                logger.error(f"Save of figure {name} failed ({fn_save}).")

        return fig, axes

    def series_per_group(
        self,
        plot_column,
        by=None,
        savefig=True,
        outputdir=".",
        naming_method=None,
        units_for_well_screen_depth="mNAP",
    ):
        """Plot time series per group.

        The default groupby is based on identical x, y coordinates, so plots unique
        time series per location.

        Parameters
        ----------
        plot_column : str
            name of column containing time series data
        by : (list of) str or (list of) array-like
            groupby parameters, default is None which sets groupby to
            columns ["x", "y"].
        savefig : bool, optional
            save figures, by default True, if False returns axes handles to plots.
        outputdir : str, optional
            path to output directory, by default the current directory (".")
        naming_method : str, optional
            method to determine file names for plots, default is None, which uses the
            following template:
                "series_by_{groupby columns}_group_{groupby values}.png".
            If a string is passed, it is interpreted as a column name and the filename
            contains the unique values from that column per group:
                "series_group_{unique values in column}.png"
            The final option includes "infer_name_monitoring_well", which attempts to
            obtain the name of the monitoring well from the final name in the group,
            by splitting the string on "-" or "_", and using the first part.
            For example, if the name is "PB001-001", the filename will be "PB001.png".
        units_for_well_screen_depth : str, optional
            if the observations are GroundWaterObs, the units of the screen depth are
            added to the legend. This keyword argument defines the units for the screen
            depth. Default is 'mNAP'.

        Returns
        -------
        axes : list
            returns list of axes handles if savefig=False
        """
        if by is None:
            by = ["x", "y"]
        gr = self._obj.groupby(by=by)

        axes = []

        for groupname, group in tqdm(
            gr, desc="Plotting series per group", total=len(gr)
        ):
            f, ax = plt.subplots(1, 1, figsize=(10, 3))
            for name, row in group.iterrows():
                if isinstance(row.obs, GroundwaterObs):
                    if units_for_well_screen_depth == "mNAP":
                        lbl = (
                            f"{name} (NAP{row['screen_top']:+.1f}"
                            f"-{row['screen_bottom']:+.1f} m)"
                        )
                    else:
                        lbl = (
                            f"{name} ({row['screen_top']:.1f}"
                            f"-{row['screen_bottom']:.1f} "
                            f"{units_for_well_screen_depth})"
                        )
                else:
                    lbl = f"{name}"
                ax.plot(
                    row.obs.index,
                    row.obs[plot_column],
                    label=lbl,
                )
            ax.legend(
                loc=(0, 1),
                frameon=False,
                ncol=min(group.index.size, 3),
                fontsize="x-small",
            )
            ax.set_ylabel(row["unit"])
            ax.grid(True)
            f.tight_layout()

            if savefig:
                if isinstance(by, list):
                    by_name = "-".join(by)
                    groupname = "-".join(groupname)
                else:
                    by_name = by
                if naming_method is None:
                    filename = f"series_by_{by_name}_group_{groupname}.png"
                elif naming_method == "infer_name_monitoring_well":
                    if "-" in name:
                        filename = f"{name.split('-')[0]}.png"
                    elif "_" in name:
                        filename = f"{name.split('_')[0]}.png"
                    else:
                        filename = f"{name}.png"
                elif isinstance(naming_method, str):
                    filename = (
                        f"series_by_{by_name}_group_"
                        f"{'-'.join(group[naming_method].unique().tolist())}.png"
                    )

                f.savefig(
                    os.path.join(outputdir, filename), bbox_inches="tight", dpi=150
                )
                plt.close(f)
            else:
                axes.append(ax)

        if not savefig:
            return axes


@accessor.register_obs_accessor("plots")
class ObsPlots:
    def __init__(self, obs):
        self._obj = obs

    def interactive_plot(
        self,
        savedir=None,
        cols=(None,),
        markers=("line",),
        p=None,
        plot_legend_names=("",),
        plot_freq=(None,),
        tmin=None,
        tmax=None,
        hoover_names=(None,),
        hoover_date_format="%Y-%m-%d",
        ylabel=None,
        plot_colors=("blue",),
        add_screen_to_legend=False,
        width=600,
        return_filename=False,
    ):
        """Create an interactive plot of the observations using bokeh.

        Todo:

        - add options for hoovers, markers, linestyle

        Parameters
        ----------
        savedir : str, optional
            directory used for the folium map and bokeh plots
        cols : tuple of str or None, optional
            the columns of the observation to plot. The first numeric column
            is used if cols is None, by default None.
        markers : list of str, optional
            type of markers that can be used for plot, 'line' and 'circle' are
            supported
        p : bokeh.plotting.figure, optional
            reference to existing figure, if p is None a new figure is created
        plot_legend_names : list of str, optional
            legend in bokeh plot
        plot_freq : list of str, optional
            bokeh plot is resampled with this frequency to reduce the size
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        hoover_names : list of str, optional
            names will be displayed together with the cols values
            when hoovering over plot
        hoover_date_format : str, optional
            date format to use when hoovering over a plot
        ylabel : str or None, optional
            label on the y-axis. If None the unit attribute of the observation
            is used.
        plot_colors : list of str, optional
            plot_colors used for the plots
        add_screen_to_legend : boolean, optional
            if True the attributes screen_top and screen_bottom
            are added to the legend name
        width : int, optional
            width of the interactive figure.
        return_filename : boolean, optional
            if True filename will be returned

        Returns
        -------
        fname_plot : str or bokeh plot
            filename of the bokeh plot or reference to bokeh plot
        """

        from bokeh.models import ColumnDataSource, HoverTool, Range1d
        from bokeh.plotting import figure, save
        from bokeh.resources import CDN

        xlim = None

        hoover_names = list(hoover_names)
        cols = list(cols)
        for i, col in enumerate(cols):
            if col is None:
                cols[i] = self._obj._get_first_numeric_col_name()
            if hoover_names[i] is None:
                hoover_names[i] = cols[i]

        # create plot dataframe
        plot_df = self._obj[tmin:tmax].copy()
        plot_df["date"] = plot_df.index.strftime(hoover_date_format)
        if plot_df.empty or plot_df[cols].isna().all().all():
            raise ValueError(
                "{} has no data between {} and {}".format(self._obj.name, tmin, tmax)
            )
        elif len(plot_df) == 1:
            markers = ["circle"] * len(cols)
            # set xlim because there is only one measurement
            obsdate = plot_df.index[0]
            xlim = Range1d(
                start=obsdate - pd.Timedelta(days=365),
                end=obsdate + pd.Timedelta(days=365),
            )

        # create plot
        if p is None:
            p = figure(
                width=width, height=int(width / 1.5), x_axis_type="datetime", title=""
            )
            if ylabel is None:
                ylabel = self._obj.unit
            p.yaxis.axis_label = ylabel

        # get x axis
        xcol = self._obj.index.name
        if xcol is None:
            xcol = "index"

        # get color
        if len(plot_colors) < len(cols):
            plot_colors = list(plot_colors) * len(cols)

        # get base for hoover tooltips
        plots = []
        tooltips = []
        tooltips.append(("date", "@date"))

        # plot multiple columns
        for i, column in enumerate(cols):
            # legend name
            if add_screen_to_legend:
                lname = "{} {} (NAP {:.2f} - {:.2f})".format(
                    plot_legend_names[i],
                    self._obj.name,
                    self._obj.screen_bottom,
                    self._obj.screen_top,
                )
            else:
                lname = "{} {}".format(plot_legend_names[i], self._obj.name)

            # resample data
            if plot_freq[i] is None:
                source = ColumnDataSource(plot_df[[column, "date"]])
            else:
                source = ColumnDataSource(
                    plot_df[[column, "date"]].resample(plot_freq[i]).first()
                )

            # plot data
            if markers[i] in ["line", "l"]:
                plots.append(
                    p.line(
                        xcol,
                        column,
                        source=source,
                        color=plot_colors[i],
                        legend_label=lname,
                        alpha=0.8,
                        muted_alpha=0.2,
                    )
                )
            elif markers[i] in ["circle", "c"]:
                plots.append(
                    p.circle(
                        xcol,
                        column,
                        source=source,
                        color=plot_colors[i],
                        legend_label=lname,
                        alpha=0.8,
                        muted_alpha=0.2,
                    )
                )
            else:
                raise NotImplementedError(
                    "marker '{}' invalid. Only line and"
                    "circle are currently available".format(markers[i])
                )

            # add columns to hoover tooltips
            tooltips_p = tooltips.copy()
            tooltips_p.append((hoover_names[i], "@{}".format(column)))
            hover = HoverTool(renderers=[plots[i]], tooltips=tooltips_p, mode="vline")
            p.add_tools(hover)

        # set xlim
        if xlim is not None:
            p.x_range = xlim

        p.legend.location = "top_left"
        p.legend.click_policy = "mute"

        # save plot
        if savedir is not None:
            if not os.path.isdir(savedir):
                os.makedirs(savedir)
            self._obj.meta["iplot_fname"] = os.path.join(
                savedir, str(self._obj.name) + ".html"
            )
            save(p, self._obj.meta["iplot_fname"], resources=CDN, title=self._obj.name)

        if return_filename:
            return self._obj.meta["iplot_fname"]
        else:
            return p



from importlib import metadata
from sys import version as os_version

__version__ = "0.16.0"


def show_versions():
    """Method to print the versions of dependencies."""
    msg = (
        f"Python version      : {os_version}\n"
        f"Numpy version       : {metadata.version('numpy')}\n"
        f"Scipy version       : {metadata.version('scipy')}\n"
        f"Pandas version      : {metadata.version('pandas')}\n"
        f"Matplotlib version  : {metadata.version('matplotlib')}\n\n"
        f"Hydropandas version : {__version__}"
    )
    return print(msg)



import logging

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

try:
    from tqdm import tqdm
except ModuleNotFoundError:
    tqdm = None

from . import accessor

logger = logging.getLogger(__name__)


def get_model_layer_z(z, zvec, left=-999, right=999):
    """Get index of model layer based on elevation.

    Assumptions:

    - the highest value in zvec is the top of model layer 0.
    - if z is equal to the bottom of a layer, the model layer above that
      layer is assigned.

    Parameters
    ----------
    z : int, float
        elevation.
    zvec : list, np.array
        elevations of model layers. shape is nlay + 1
    left : int, optional
        if z is below the lowest value in zvec, this value is returned.
        The default is -999.
    right : TYPE, optional
        if z is above the highest value in zvec, this value is returned.
        The default is 999.

    Returns
    -------
    int : int
        model layer

    Examples
    --------
    >>> zvec = [0, -10, -20, -30]
    >>> get_model_layer_z(-5, zvec)
    0

    >>> get_model_layer_z(-25, zvec)
    2

    >>> get_model_layer_z(-50, zvec)
    -999

    >>> get_model_layer_z(100, zvec)
    999

    >>> get_model_layer_z(-20, zvec)
    1
    """
    # make sure zvec is descending
    zvec = np.sort(zvec)[::-1]
    if z in zvec:
        z += 1e-10
    lay = int(
        np.interp(z, zvec[::-1], np.arange(len(zvec))[::-1], left=left, right=right)
    )

    return lay


def check_if_var_is_invalid(var):
    if var is None:
        return True
    elif np.isnan(var):
        return True

    return False


def get_modellayer_from_screen_depth(ftop, fbot, zvec, left=-999, right=999):
    """


    Parameters
    ----------
    ftop : int or float
        top of screen.
    fbot : int or float
        bottom of screen, has to be lower than ftop.
    zvec : list or np.array
        elevations of the modellayers at the location of the tube.
    left : int, optional
        value to return if tube screen is below the modellayers.
        The default is -999.
    right : int, optional
        value to return if tube screen is above the modellayers.
        The default is-999.

    Raises
    ------
    ValueError
        raised if something unexpected happens.

    Returns
    -------
    int or np.nan
        modellayer.

    examples
    --------
    >>> zvec = [0, -10, -20, -30, -40]
    >>> get_modellayer_from_screen_depth(-5, -7, zvec)
    0

    >>> get_modellayer_from_screen_depth(-25, -27, zvec)
    2

    >>> get_modellayer_from_screen_depth(-15, -27, zvec)
    2

    >>> get_modellayer_from_screen_depth(-5, -27, zvec)
    1

    >>> get_modellayer_from_screen_depth(-5, -37, zvec)
    1

    >>> get_modellayer_from_screen_depth(15, -5, zvec)
    0

    >>> get_modellayer_from_screen_depth(15, 5, zvec)
    999

    >>> get_modellayer_from_screen_depth(-55, -65, zvec)
    -999

    >>> get_modellayer_from_screen_depth(15, -65, zvec)
    nan

    >>> get_modellayer_from_screen_depth(None, -7, zvec)
    0

    >>> get_modellayer_from_screen_depth(None, None, zvec)
    nan

    """
    if isinstance(zvec, np.ndarray):
        if np.isnan(zvec).all():
            logger.warning("vertical datum invalid returning nan")
            return np.nan

    zvec = np.sort(zvec)[::-1]
    ftop_invalid = check_if_var_is_invalid(ftop)
    fbot_invalid = check_if_var_is_invalid(fbot)
    if ftop_invalid and fbot_invalid:
        logger.error("- values for ftop and fbot are invalid!")
        return np.nan
    elif fbot_invalid:
        lay_ftop = get_model_layer_z(ftop, zvec, left=left, right=right)
        logger.error(f"- fbot invalid. selected based on ftop: {lay_ftop}")
        return lay_ftop
    elif ftop_invalid:
        lay_fbot = get_model_layer_z(fbot, zvec, left=left, right=right)
        logger.error(f"- ftop invalid. selected based on fbot: {lay_fbot}")
        return lay_fbot

    if ftop < fbot:
        logger.warning("- tube screen top below tube screen bot, switching top and bot")
        fbot, ftop = ftop, fbot

    lay_ftop = get_model_layer_z(ftop, zvec, left=left, right=right)
    lay_fbot = get_model_layer_z(fbot, zvec, left=left, right=right)

    # Piezometer in layer in which majority of screen is located
    if lay_fbot == lay_ftop:
        logger.debug(f"- selected layer {lay_fbot}")
        return lay_fbot

    else:
        if lay_fbot == left and lay_ftop == right:
            logger.debug("- tube screen spans all layers. return nan")
            return np.nan
        elif lay_ftop == right:
            logger.debug(
                f"- tube screen top higher than top layer. selected layer {lay_fbot}"
            )
            return lay_fbot

        elif lay_fbot == left:
            logger.debug(
                f"- tube screen bot lower than bottom layer. selected layer {lay_ftop}"
            )
            return lay_ftop

        logger.debug(
            f"- tube screen crosses layer boundary:\n  - layers: {lay_ftop}, {lay_fbot}"
        )

        logger.debug(
            f"- tube screen spans {lay_fbot - lay_ftop + 1} layers."
            " checking length per layer\n"
            "  - length per layer:"
        )

        # check which layer has the biggest length of the tube screen
        length_layers = np.zeros(lay_fbot - lay_ftop + 1)
        for i in range(len(length_layers)):
            if i == 0:
                length_layers[i] = ftop - zvec[lay_ftop + 1]
            elif (i + 1) == len(length_layers):
                length_layers[i] = zvec[lay_ftop + i] - fbot
            else:
                length_layers[i] = zvec[lay_ftop + i] - zvec[lay_ftop + 1 + i]

            logger.debug(f"    - lay {lay_ftop + i}: {length_layers[i]:.2f}")

        # choose layer with biggest length
        rel_layer = np.argmax(length_layers)
        lay_out = lay_ftop + rel_layer
        logger.debug(f"  - selected layer: {lay_out}")
        return lay_out


def get_zvec(x, y, gwf=None, ds=None):
    """get a list with the vertical layer boundaries at a point in the model.

    Parameters
    ----------
    x : int or float
        x coordinate.
    y : int or float
        y coordinate.
    gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
        modflow model with top and bottoms
    ds : xarray.Dataset
        xarray Dataset typically created in nlmod. Must have
        variables 'top' and 'botm'.

    Raises
    ------
    NotImplementedError
        not all grid types are supported yet.

    Returns
    -------
    zvec : list
        list of vertical layer boundaries. length is nlay + 1.
    """
    import flopy
    from shapely.geometry import Point

    if gwf and not ds:
        ix = flopy.utils.GridIntersect(gwf.modelgrid, method="vertex")
        if gwf.modelgrid.grid_type == "structured":
            res = ix.intersect(Point(x, y))
            if len(res) > 0:
                r, c = res["cellids"][0]
                zvec = np.array(
                    [gwf.modelgrid.top[r, c]]
                    + [gwf.modelgrid.botm[i, r, c] for i in range(gwf.modelgrid.nlay)]
                )
            else:
                print("Point is not within model extent! Returning NaN.")
                zvec = np.nan
        elif gwf.modelgrid.grid_type == "vertex":
            res = ix.intersect([(x, y)], shapetype="point")
            idx = res["cellids"].astype(int)[0]
            if len(res) > 0:
                zvec = [gwf.modelgrid.top[idx]] + [
                    gwf.modelgrid.botm[i, idx] for i in range(gwf.modelgrid.nlay)
                ]
            else:
                print("Point is not within model extent! Returning NaN.")
                zvec = np.nan
        else:
            raise NotImplementedError(
                f"gridtype {gwf.modelgrid.grid_type} not (yet) implemented"
            )
    elif ds and not gwf:
        # assuming modflow type definition with 1 top and N botms
        # first check if point is within the extent
        if "angrot" in ds.attrs and ds.attrs["angrot"] != 0.0:
            import nlmod

            pol = nlmod.dims.get_extent_polygon(ds)
            if not Point(x, y).within(pol):
                return np.nan
        else:
            xmin, xmax, ymin, ymax = ds.attrs["extent"]
            if (x < xmin) or (x > xmax) or (y < ymin) or (y > ymax):
                return np.nan

        if ds.gridtype == "vertex":
            import nlmod

            cid = nlmod.dims.xy_to_icell2d((x, y), ds)
            sel = ds.sel(icell2d=cid)
            zvec = np.concatenate(([sel["top"].values], sel["botm"].values))
            mask = np.isnan(zvec)
            idx = np.where(~mask, np.arange(mask.size), 0)
            np.maximum.accumulate(idx, axis=0, out=idx)
            zvec[mask] = zvec[idx[mask]]
        else:
            sel = ds.sel(x=x, y=y, method="nearest")
            first_notna = np.nonzero(np.isfinite(np.atleast_1d(sel["top"].values)))[0][
                0
            ]
            if sel["top"].values.shape == tuple():
                top = np.atleast_1d(sel["top"].values)
            else:
                top = np.atleast_1d(sel["top"].values[[first_notna]])
            zvec = np.concatenate([top, sel["botm"].values])
            mask = np.isnan(zvec)
            idx = np.where(~mask, np.arange(mask.size), 0)
            np.maximum.accumulate(idx, axis=0, out=idx)
            zvec[mask] = zvec[idx[mask]]
    else:
        raise ValueError("Pass one of 'gwf' or 'ds'!")

    return zvec


@accessor.register_obscollection_accessor("gwobs")
class GwObsAccessor:
    def __init__(self, oc_obj):
        self._obj = oc_obj

    def set_tube_nr(
        self, radius=1, xcol="x", ycol="y", if_exists="error", add_to_meta=False
    ):
        """Set the tube numbers based on the location of the observations.

        The value of the tube number is set:

        - In the ObsCollection dataframe
        - As the attribute of an Obs object
        - optionally: In the meta dictionary of the Obs object (only if add_to_meta is
          True)

        This method is useful for groundwater observations. If two or more
        observation points are close to each other they will be seen as one
        monitoring_well with multiple tubes. The tube_nr is based on the
        'screen_bottom' attribute of the observations in such a way that
        the deepest tube has the highest tube number.

        Parameters
        ----------
        radius : int, optional
            max distance between two observations to be seen as one location,
            by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a tube_nr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the tube_nr is added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column tube_nr exists and if_exists='error' an error is
            raised
        """

        # check if column exists in obscollection
        if "tube_nr" in self._obj.columns:
            # set type to numeric
            if not is_numeric_dtype(self._obj["tube_nr"]):
                self._obj["tube_nr"] = pd.to_numeric(
                    self._obj["tube_nr"], errors="coerce"
                )

            # check if name should be replaced
            if if_exists == "error":
                raise RuntimeError(
                    "the column 'tube_nr' already exist, set"
                    "if_exists='replace' to replace the current values"
                )
            elif if_exists == "replace":
                self._obj["tube_nr"] = np.nan
        else:
            self._obj["tube_nr"] = np.nan

        # apply tube numbers to tubes that are close to eachother
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, "tube_nr"]):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_tubes = np.sqrt(
                    (self._obj[xcol] - x) ** 2 + (self._obj[ycol] - y) ** 2
                )
                dup_x = self._obj.loc[distance_to_other_tubes < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(
                        name, "tube_nr", 1, add_to_meta=add_to_meta
                    )
                else:
                    dup_x2 = dup_x.sort_values("screen_bottom", ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        self._obj._set_metadata_value(
                            pb_dub, "tube_nr", i + 1, add_to_meta=add_to_meta
                        )

    def set_tube_nr_location(
        self,
        loc_col,
        radius=1,
        xcol="x",
        ycol="y",
        if_exists="error",
        add_to_meta=False,
    ):
        """This method sets the tube_nr and location attributes of a GroundwaterObs
        based on the location and screen depth of the observations.

        The value of the tube_nr and the location are set:
        - in the ObsCollection dataframe
        - as the attribute of an Obs object
        - optional: in the meta dictionary of the Obs object (only if add_to_meta is
          True)

        If two or more observations are close to each other they will be seen
        as one location with multiple tubes. The tube_nr is based on the
        'screen_bottom' attribute of the observations in such a way that
        the deepest tube has the highest tube number. The location name is
        based on the name in the loc_col of the screen with the lowest
        tube_nr.

        Parameters
        ----------
        loc_col : str
            the column name with the names to use for the location
        radius : int, optional
            max distance between two observations to be seen as one location,
            by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a tube_nr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the tube_nr and location are added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column tube_nr exists and if_exists='error' an error
            is raised
        """

        # check if columns exists in obscollection
        if "tube_nr" in self._obj.columns or "location" in self._obj.columns:
            if if_exists == "error":
                raise RuntimeError(
                    "the column 'tube_nr or location' already exist, set"
                    "if_exists='replace' to replace the current values"
                )
            elif if_exists == "replace":
                self._obj["tube_nr"] = np.nan
                self._obj["location"] = np.nan
        else:
            self._obj["tube_nr"] = np.nan
            self._obj["location"] = np.nan

        # apply tube numbers to tubes that are close to eachother
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, "tube_nr"]):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_tubes = np.sqrt(
                    (self._obj[xcol] - x) ** 2 + (self._obj[ycol] - y) ** 2
                )
                dup_x = self._obj.loc[distance_to_other_tubes < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(
                        name, "tube_nr", 1, add_to_meta=add_to_meta
                    )
                    location = self._obj.loc[name, loc_col]
                    self._obj._set_metadata_value(
                        name,
                        "location",
                        location,
                        add_to_meta=add_to_meta,
                    )
                else:
                    dup_x2 = dup_x.sort_values("screen_bottom", ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        if i == 0:
                            location = self._obj.loc[pb_dub, loc_col]
                        self._obj._set_metadata_value(
                            pb_dub, "tube_nr", i + 1, add_to_meta=add_to_meta
                        )
                        self._obj._set_metadata_value(
                            pb_dub,
                            "location",
                            location,
                            add_to_meta=add_to_meta,
                        )

    def get_modellayers(self, gwf=None, ds=None):
        """Get the modellayer per observation. The layers can be obtained from
        the modflow model or can be defined in zgr.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        ds : xarray.Dataset
            xarray Dataset with with top and bottoms, must have
            dimensions 'x' and 'y' and variables 'top' and 'bot'

        Returns
        -------
        pd.Series with the modellayers of each observation
        """
        msg = (
            "the get_modellayers method is deprecated, please use the nlmod "
            "function: nlmod.layer.get_modellayers_screens"
        )
        logger.warning(msg)

        modellayers = []
        for o in self._obj.obs.values:
            logger.debug("-" * 10 + f"\n {o.name}:")

            modellayers.append(o.gwobs.get_modellayer_modflow(gwf=gwf, ds=ds))

        modellayers = pd.Series(
            index=self._obj.index, data=modellayers, name="modellayer"
        )

        return modellayers

    def get_regis_layers(self):
        """Get the regis layer per observation.

        Parameters
        ----------


        Returns
        -------
        pd.Series with the names of the regis layer of each observation
        """
        if tqdm is not None:
            tqdm.pandas()
            return self._obj.obs.progress_apply(lambda o: o.gwobs.get_regis_layer())
        else:
            return self._obj.obs.apply(lambda o: o.gwobs.get_regis_layer())


@accessor.register_obs_accessor("gwobs")
class GeoAccessorObs:
    def __init__(self, obs):
        self._obj = obs

    def get_modellayer_modflow(self, gwf=None, ds=None, left=-999, right=999):
        """Add modellayer to meta dictionary.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        ds : xarray.Dataset
            xarray Dataset with with top and bottoms, must have
            dimensions 'x' and 'y' and variables 'top' and 'bot'

        Returns
        -------
        int
            modellayer
        """
        if np.isnan(self._obj.x) or np.isnan(self._obj.y):
            logger.warning("x or y coordinate is NaN, returning NaN")
            return np.nan

        zvec = get_zvec(self._obj.x, self._obj.y, gwf=gwf, ds=ds)

        if np.all(np.isnan(zvec)):
            return np.nan
        else:
            modellayer = get_modellayer_from_screen_depth(
                self._obj.screen_top,
                self._obj.screen_bottom,
                zvec,
                left=left,
                right=right,
            )
            return modellayer

    def get_regis_layer(self):
        """find the name of the REGIS layer based on the tube screen depth.

        Parameters
        ----------


        Returns
        -------
        str
            name of REGIS layer
        """
        import xarray as xr

        if np.isnan(self._obj.screen_top) or np.isnan(self._obj.screen_bottom):
            return "nan"

        # connect to regis netcdf
        regis_url = "https://dinodata.nl/opendap/REGIS/REGIS.nc"
        regis_ds = xr.open_dataset(regis_url, engine="netcdf4", decode_times=False)

        # rename layer in regis netcdf
        regis_ds = regis_ds.rename({"layer": "layer_old"})
        regis_ds.coords["layer"] = regis_ds.layer_old.astype(str)
        regis_ds = regis_ds.swap_dims({"layer_old": "layer"})

        # get zvec regis netcdf
        z = (
            regis_ds["bottom"]
            .sel(x=self._obj.x, y=self._obj.y, method="nearest")
            .dropna(dim="layer")
        )
        ztop = regis_ds["top"].sel(x=self._obj.x, y=self._obj.y, method="nearest").max()
        zvec = np.concatenate([[ztop.values], z.values])

        # get index of regis model layer
        layer_i = get_modellayer_from_screen_depth(
            self._obj.screen_top,
            self._obj.screen_bottom,
            zvec,
            left=-999,
            right=999,
        )

        if layer_i == 999:
            return "above"
        elif layer_i == -999:
            return "below"

        return str(z.layer[layer_i].values)



"""
This module is all about data is to be submitted to
https://www.bronhouderportaal-bro.nl, i.e. XMLs of GMWs that are newly
installed by a fieldwork company
"""

import logging
import pathlib
import xml.etree.ElementTree

import pandas as pd

logger = logging.getLogger(__name__)


def get_tube_nrs_from_xml(tree, ns):
    """
    Get tube numbers from xml data

    Parameters
    ----------
    tree : xml.etree.ElementTree.ElementTree
        The XML input.
    ns : dictonary
        Namespaces in XML input

    Returns
    -------
    all_tube_nrs : list
        The numbers of all filters in the xml data.

    """
    # get numbers of individual filters from XML file
    all_tube_nrs = []
    tubes = tree.findall(
        "isgmw:sourceDocument//isgmw:GMW_Construction//isgmw:monitoringTube", ns
    )
    for tube in tubes:
        all_tube_nrs.append(int(tube.find("isgmw:tubeNumber", ns).text))

    return all_tube_nrs


def get_obs_list_from_dir(
    dirname,
    ObsClass,
    full_meta=False,
):
    """get a list of gmw observations within a dirname.

    Parameters
    ----------
    dirname : str
        name of directory with XML files
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    full_meta : bool, optional
        process not only the standard metadata to ObsCollection


    Returns
    -------
    obs_list : list
        all observation wells

    """
    ns = {
        "brocommon": "http://www.broservices.nl/xsd/brocommon/3.0",
        "isgmw": "http://www.broservices.nl/xsd/isgmw/1.1",
        "gmwcommon": "http://www.broservices.nl/xsd/gmwcommon/1.1",
        "d5p1": "http://www.opengis.net/gml/3.2",
    }

    obs_list = []

    # loop over files
    for path_xml in pathlib.Path(dirname).glob("*.xml"):
        # open xml and get number of monitoring tubes
        tree = xml.etree.ElementTree.parse(path_xml)
        # get number of tubes from XML file
        ntubes_from_tree = int(
            tree.find(
                "isgmw:sourceDocument//"
                "isgmw:GMW_Construction//"
                "isgmw:numberOfMonitoringTubes",
                ns,
            ).text
        )
        all_tube_nrs = get_tube_nrs_from_xml(tree, ns)
        if len(all_tube_nrs) != ntubes_from_tree:
            logger.warning(
                f"{path_xml} numberOfMonitoringTubes suggests that "
                f"{ntubes_from_tree} filters are present, "
                f"however meta data for {len(all_tube_nrs)} filters is "
                f"in the xml. Meta data for {len(all_tube_nrs)} filters "
                "is processed."
            )

        for tube_nr in all_tube_nrs:
            # append each tube to the obs_list
            o = ObsClass.from_bronhouderportaal_bro(
                path_xml,
                tube_nr=tube_nr,
                full_meta=full_meta,
            )
            obs_list.append(o)

    return obs_list


def get_metadata_from_gmw(path_xml, tube_nr, full_meta=False):
    """get metadata for a groundwater monitoring well.
    standard: coordinates, ground_level, tube_top and tube screen

    Function is based upon get_metadata_from_gmw in bro.py, changes are made because:
        - bro_id is not available, because data is not yet imported to bro
        - XML layout of input is slightly different (example of Fugro is used)


    Parameters
    ----------
    path_xml : pathlib.WindowsPath
        path of groundwater monitoring well XML file.
    tube_nr : int
        tube number you want metadata for.

    Raises
    ------
    ValueError
        if bro_id is invalid.
    TypeError
        if tube_nr is not an int

    Returns
    -------
    meta : dict
        dictionary with metadata.

    """
    if not isinstance(path_xml, pathlib.WindowsPath):
        try:
            path_xml = pathlib.Path(path_xml)
        except ValueError:
            print(f"invalid path_xml {path_xml}")

    # open meta dictonary
    meta = {
        "filename": path_xml.name,
        "tube_nr": tube_nr,
        "source": "bronhouderportaal-bro",
    }

    # open xml file
    tree = xml.etree.ElementTree.parse(path_xml)
    ns = {
        "brocommon": "http://www.broservices.nl/xsd/brocommon/3.0",
        "isgmw": "http://www.broservices.nl/xsd/isgmw/1.1",
        "gmwcommon": "http://www.broservices.nl/xsd/gmwcommon/1.1",
        "d5p1": "http://www.opengis.net/gml/3.2",
    }

    # is tube_nr valid
    all_tube_nrs = get_tube_nrs_from_xml(tree, ns)
    if tube_nr not in all_tube_nrs:
        raise ValueError(
            f"Request is to process tube_nr {tube_nr}, that number is not "
            f"in the XML data. Valid tube_nrs are {all_tube_nrs} "
        )

    # name of well via requestReference
    # bro_id is not used, because well is not yet imported in Broloket
    if tree.find("brocommon:requestReference", ns) is not None:
        name = tree.find("brocommon:requestReference", ns).text
    else:
        name = path_xml.stem  # filname without extention
    meta["name"] = f"{name}-{tube_nr}"
    meta["location"] = name

    if full_meta:
        if tree.find("brocommon:deliveryAccountableParty", ns) is not None:
            # deliveryAccountableParty is KvK nummer
            meta["deliveryAccountableParty"] = tree.find(
                "brocommon:deliveryAccountableParty", ns
            ).text
        if tree.find("brocommon:qualityRegime", ns) is not None:
            meta["qualityRegime"] = tree.find("brocommon:qualityRegime", ns).text

    # open sourceDocument part
    sourceDocument = tree.find("isgmw:sourceDocument", ns)

    # GMW_Construction element
    GMW_c = sourceDocument.find("isgmw:GMW_Construction", ns)

    if full_meta:
        if GMW_c.find("isgmw:objectIdAccountableParty", ns) is not None:
            meta["objectIdAccountableParty"] = GMW_c.find(
                "isgmw:objectIdAccountableParty", ns
            ).text
        if GMW_c.find("isgmw:deliveryContext", ns) is not None:
            meta["deliveryContext"] = GMW_c.find("isgmw:deliveryContext", ns).text
        if GMW_c.find("isgmw:constructionStandard", ns) is not None:
            meta["constructionStandard"] = GMW_c.find(
                "isgmw:constructionStandard", ns
            ).text
        if GMW_c.find("isgmw:initialFunction", ns) is not None:
            meta["initialFunction"] = GMW_c.find("isgmw:initialFunction", ns).text
        if GMW_c.find("isgmw:groundLevelStable", ns) is not None:
            meta["groundLevelStable"] = GMW_c.find("isgmw:groundLevelStable", ns).text
        if GMW_c.find("isgmw:owner", ns) is not None:
            meta["owner"] = GMW_c.find("isgmw:owner", ns).text
        if GMW_c.find("isgmw:wellHeadProtector", ns) is not None:
            meta["wellHeadProtector"] = GMW_c.find("isgmw:wellHeadProtector", ns).text
        if GMW_c.find("isgmw:wellConstructionDate", ns) is not None:
            xml_text = GMW_c.find("isgmw:wellConstructionDate//brocommon:date", ns).text
            meta["wellConstructionDate"] = pd.Timestamp(xml_text)

    # x and y
    xy = GMW_c.find("isgmw:deliveredLocation//gmwcommon:location//d5p1:pos", ns)
    meta["x"], meta["y"] = [float(val) for val in xy.text.split()]
    if full_meta:
        if (
            GMW_c.find(
                "isgmw:deliveredLocation//gmwcommon:horizontalPositioningMethod", ns
            )
            is not None
        ):
            meta["horizontalPositioningMethod"] = GMW_c.find(
                "isgmw:deliveredLocation//gmwcommon:horizontalPositioningMethod", ns
            ).text
        if (
            GMW_c.find(
                "isgmw:deliveredVerticalPosition//"
                "gmwcommon:localVerticalReferencePoint",
                ns,
            )
            is not None
        ):
            meta["localVerticalReferencePoint"] = GMW_c.find(
                "isgmw:deliveredVerticalPosition//"
                "gmwcommon:localVerticalReferencePoint",
                ns,
            ).text
        if (
            GMW_c.find("isgmw:deliveredVerticalPosition//gmwcommon:offset", ns)
            is not None
        ):
            meta["deliveredVerticalPosition_offset"] = float(
                GMW_c.find("isgmw:deliveredVerticalPosition//gmwcommon:offset", ns).text
            )
        if (
            GMW_c.find(
                "isgmw:deliveredVerticalPosition//"
                "gmwcommon:groundLevelPositioningMethod",
                ns,
            )
            is not None
        ):
            meta["groundLevelPositioningMethod"] = GMW_c.find(
                "isgmw:deliveredVerticalPosition//"
                "gmwcommon:groundLevelPositioningMethod",
                ns,
            ).text

    # ground_level
    glp_xml = GMW_c.find(
        "isgmw:deliveredVerticalPosition//gmwcommon:groundLevelPosition", ns
    )
    vert_datum = GMW_c.find(
        "isgmw:deliveredVerticalPosition//gmwcommon:verticalDatum", ns
    ).text
    meta["unit"] = glp_xml.attrib["uom"] + " " + vert_datum
    if glp_xml.attrib["uom"].lower() != "m":
        logger.info(
            f"groundlevel unit is unexpected {glp_xml.attrib['uom']}, m is expected"
        )
    if vert_datum.lower() != "nap":
        logger.info(f"datum has unexpected value {vert_datum}, NAP is expected")
    meta["ground_level"] = float(glp_xml.text)

    # tube properties
    tubes = GMW_c.findall("isgmw:monitoringTube", ns)
    for tube in tubes:
        # open only requested tube
        if int(tube.find("isgmw:tubeNumber", ns).text) == tube_nr:
            break

    if full_meta:
        if tube.find("isgmw:tubeType", ns) is not None:
            meta["tubeType"] = tube.find("isgmw:tubeType", ns).text
        if tube.find("isgmw:artesianWellCapPresent", ns) is not None:
            meta["artesianWellCapPresent"] = tube.find(
                "isgmw:artesianWellCapPresent", ns
            ).text
        if tube.find("isgmw:sedimentSumpPresent", ns) is not None:
            meta["sedimentSumpPresent"] = tube.find(
                "isgmw:sedimentSumpPresent", ns
            ).text
        if tube.find("isgmw:numberOfGeoOhmCables", ns) is not None:
            meta["numberOfGeoOhmCables"] = int(
                tube.find("isgmw:numberOfGeoOhmCables", ns).text
            )
        if tube.find("isgmw:tubeTopDiameter", ns) is not None:
            meta["tubeTopDiameter"] = float(tube.find("isgmw:tubeTopDiameter", ns).text)
            meta["tubeTopDiameter_unit"] = tube.find(
                "isgmw:tubeTopDiameter", ns
            ).attrib["uom"]
        if tube.find("isgmw:variableDiameter", ns) is not None:
            meta["variableDiameter"] = tube.find("isgmw:variableDiameter", ns).text
        if tube.find("isgmw:tubeStatus", ns) is not None:
            meta["tubeStatus"] = tube.find("isgmw:tubeStatus", ns).text
        if tube.find("isgmw:tubeTopPositioningMethod", ns) is not None:
            meta["tubeTopPositioningMethod"] = tube.find(
                "isgmw:tubeTopPositioningMethod", ns
            ).text
        if (
            tube.find("isgmw:materialUsed//gmwcommon:tubePackingMaterial", ns)
            is not None
        ):
            meta["tubePackingMaterial"] = tube.find(
                "isgmw:materialUsed//gmwcommon:tubePackingMaterial", ns
            ).text
        if tube.find("isgmw:materialUsed//gmwcommon:tubeMaterial", ns) is not None:
            meta["tubeMaterial"] = tube.find(
                "isgmw:materialUsed//gmwcommon:tubeMaterial", ns
            ).text
        if tube.find("isgmw:materialUsed//gmwcommon:glue", ns) is not None:
            meta["tubeGlue"] = tube.find("isgmw:materialUsed//gmwcommon:glue", ns).text
        if tube.find("isgmw:screen//isgmw:sockMaterial", ns) is not None:
            meta["sockMaterial"] = tube.find(
                "isgmw:screen//isgmw:sockMaterial", ns
            ).text

    # tube_top
    tube_top_xml = tube.find("isgmw:tubeTopPosition", ns)
    meta["tube_top"] = float(tube_top_xml.text)
    tube_top_unit = tube_top_xml.attrib["uom"]
    if tube_top_unit != "m":
        logger.info(f"tube_top unit is unexpected {tube_top_unit}, m expected")

    # some lenghts to calculate screen levels
    screenLength_xml = tube.find("isgmw:screen//isgmw:screenLength", ns)
    screenLength = float(screenLength_xml.text)
    if full_meta:
        meta["screenLength"] = screenLength
    screenLength_unit = screenLength_xml.attrib["uom"]
    if screenLength_unit != "m":
        logger.info(
            f"screenLength unit is unexpected {screenLength.attrib['uom']},m expected"
        )

    plainTubePartLength_xml = tube.find(
        "isgmw:plainTubePart//gmwcommon:plainTubePartLength", ns
    )
    plainTubePartLength = float(plainTubePartLength_xml.text)
    if full_meta:
        meta["plainTubePartLength"] = plainTubePartLength
    plainTubePartLength_unit = plainTubePartLength_xml.attrib["uom"]
    if plainTubePartLength_unit != "m":
        logger.info(
            "plainTubePartLength unit is unexpected"
            f"{plainTubePartLength_unit}, m expected"
        )

    if tube.find("isgmw:sedimentSumpPresent", ns).text.lower() in ["ja", "yes"]:
        sedimentSumpLength_xml = tube.find(
            "isgmw:sedimentSump//gmwcommon:sedimentSumpLength", ns
        )
        sedimentSumpLength = float(sedimentSumpLength_xml.text)
        sedimentSumpLength_unit = sedimentSumpLength_xml.attrib["uom"]
        if sedimentSumpLength_unit != "m":
            logger.info(
                "sedimentSumpLength unit is unexpected"
                f"{sedimentSumpLength_unit}, m expected"
            )

        if full_meta:
            meta["sedimentSumpLength"] = sedimentSumpLength

    if (
        ("tube_top" in meta)
        and ("plainTubePartLength" in locals())
        and (tube_top_unit == plainTubePartLength_unit)
    ):
        meta["screen_top"] = round(meta["tube_top"] - plainTubePartLength, 2)
    if (
        ("screen_top" in meta)
        and ("screenLength" in locals())
        and (plainTubePartLength_unit == screenLength_unit)
    ):
        meta["screen_bottom"] = round(meta["screen_top"] - screenLength, 2)
    if (
        ("screen_bottom" in meta)
        and ("sedimentSumpLength" in locals())
        and (screenLength_unit == sedimentSumpLength_unit)
    ):
        meta["tube_bottom"] = round(meta["screen_bottom"] - sedimentSumpLength, 2)

    meta["metadata_available"] = True

    return meta



"""
https://www.bro-productomgeving.nl/bpo/latest/informatie-voor-softwareleveranciers/url-s-publieke-rest-services

function levels:
1. get_obs_list_from_gmn: list of observations from grondwatermonitoring net
    2. get_bro_groundwater: single observation
        3. measurements_from_gld:
        4. get_metadata_from_gmw:
        5. get_gld_ids_from_gmw:

"""

import json
import logging
import xml.etree.ElementTree
from functools import lru_cache

import numpy as np
import pandas as pd
import requests
from pyproj import Proj, Transformer
from requests.adapters import HTTPAdapter, Retry
from tqdm import tqdm

from ..rcparams import rcParams
from ..util import EPSG_28992

logger = logging.getLogger(__name__)


def get_obs_list_from_gmn(bro_id, ObsClass, only_metadata=False, keep_all_obs=True):
    """get a list of observation from a groundwater monitoring network.

    Parameters
    ----------
    bro_id : str
        starts with 'GMN' e.g. 'GMN000000000163'.
    ObsClass : type
        class of the observations, so far only GroundwaterObs is supported
    only_metadata : bool, optional
        if True download only metadata, significantly faster. The default
        is False.
    keep_all_obs : boolean, optional
        add all observation points to the collection, even without
        measurements

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    obs_list : list
        list with observation objects.
    meta : dict
        metadata of the groundwater monitoring net.

    """

    if not bro_id.startswith("GMN"):
        raise ValueError("bro id should start with GMN")

    url = f"https://publiek.broservices.nl/gm/gmn/v1/objects/{bro_id}"
    req = requests.get(url)

    if req.status_code > 200:
        logger.error(
            "could not get monitoring wells, this groundwater monitoring net is"
            "probably too big. Try a smaller extent"
        )
        req.raise_for_status()

    ns = {"xmlns": "http://www.broservices.nl/xsd/dsgmn/1.0"}

    tree = xml.etree.ElementTree.fromstring(req.text)
    gmn = tree.find(".//xmlns:GMN_PO", ns)
    gmws = gmn.findall("xmlns:measuringPoint", ns)

    logger.info(f"{len(gmws)} groundwater monitoring wells within groundwater meetnet")

    obs_list = []
    for gmw in tqdm(gmws):
        tags = ["MeasuringPoint", "monitoringTube", "GroundwaterMonitoringTube"]
        tube = gmw.find(f".//xmlns:{'//xmlns:'.join(tags)}", ns)
        gmw_id = tube.find("xmlns:broId", ns).text
        tube_nr = int(tube.find("xmlns:tubeNumber", ns).text)

        o = ObsClass.from_bro(
            bro_id=gmw_id, tube_nr=tube_nr, only_metadata=only_metadata
        )
        if o.empty:
            logger.debug(
                f"no measurements found for gmw_id {gmw_id} and tube number {tube_nr}"
            )
            if keep_all_obs:
                obs_list.append(o)
        else:
            obs_list.append(o)
        obs_list.append(o)

    meta = {}
    meta["name"] = gmn.find("xmlns:name", ns).text
    meta["doel"] = gmn.find("xmlns:monitoringPurpose", ns).text
    meta["bro_id"] = bro_id

    return obs_list, meta


def get_bro_groundwater(bro_id, tube_nr=None, only_metadata=False, **kwargs):
    """get bro groundwater measurement from a GLD id or a GMW id with a
    filter number.


    Parameters
    ----------
    bro_id : str
        starts with 'GLD' or 'GMW' e.g. 'GLD000000012893'.
    tube_nr : int or None, optional
        tube number, required if bro_id starts with 'GMW'. The default is
        None.
    only_metadata : bool, optional
        if True download only metadata, significantly faster. The default
        is False.
    **kwargs :
        passes to measurements_from_gld.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    dataframe
        measurements.
    dictionary
        metadata.

    """
    logger.debug(f"reading bro_id {bro_id}")

    if bro_id.startswith("GLD"):
        if only_metadata:
            raise ValueError("cannot get metadata from gld id")
        return measurements_from_gld(bro_id, **kwargs)

    elif bro_id.startswith("GMW"):
        if tube_nr is None:
            raise ValueError("if bro_id is GMW a tube_nr should be specified")

        meta = get_metadata_from_gmw(bro_id, tube_nr)
        gld_ids = get_gld_ids_from_gmw(bro_id, tube_nr)

        if gld_ids is None:
            meta["name"] = f"{bro_id}_{tube_nr}"
            only_metadata = True  # cannot get time series without gld id
        else:
            meta["name"] = f"{bro_id}_{tube_nr}"
            meta["gld_ids"] = gld_ids

        if only_metadata:
            empty_df = pd.DataFrame()
            return empty_df, meta

        dfl = []
        for i, gld_id in enumerate(gld_ids):
            df, meta_new = measurements_from_gld(gld_id, **kwargs)
            meta.update(meta_new)
            dfl.append(df)
        df = pd.concat(dfl, axis=0).sort_index()

        return df, meta


def get_gld_ids_from_gmw(bro_id, tube_nr):
    """get bro_ids of multiple grondwaterstandendossier (gld) from a bro_id of a
    grondwatermonitoringsput (gmw).

    Parameters
    ----------
    bro_id : str
        starts with 'GLD' or 'GMW' e.g. 'GMW000000036287'.
    tube_nr : int
        tube number.

    Raises
    ------
    ValueError
        DESCRIPTION.
    RuntimeError
        DESCRIPTION.

    Returns
    -------
    list of str or None
        bro_ids of a grondwaterstandonderzoek (gld).

    """
    if not bro_id.startswith("GMW"):
        raise ValueError("bro id should start with GMW")

    url = f"https://publiek.broservices.nl/gm/v1/gmw-relations/{bro_id}"
    req = requests.get(url)

    if req.status_code > 200:
        logger.error(f"could not get gld ids {req.text}")
        req.raise_for_status()

    d = json.loads(req.text)

    if len(d["monitoringTubeReferences"]) == 0:
        logger.debug(
            f"no groundwater level dossier for {bro_id} and tube number {tube_nr}"
        )
        return None

    for tube in d["monitoringTubeReferences"]:
        if tube["tubeNumber"] == tube_nr:
            if len(tube["gldReferences"]) == 0:
                logger.debug(
                    f"no groundwater level dossier for {bro_id} and tube number"
                    f"{tube_nr}"
                )
                return None
            else:
                return [gldref["broId"] for gldref in tube["gldReferences"]]


def measurements_from_gld(
    bro_id, tmin=None, tmax=None, to_wintertime=True, drop_duplicate_times=True
):
    """get measurements and metadata from a grondwaterstandonderzoek (gld)
    bro_id


    Parameters
    ----------
    bro_id : str
        e.g. 'GLD000000012893'.
    tmin : str or None, optional
        start date in format YYYY-MM-DD
    tmax : str or None, optional
        end date in format YYYY-MM-DD
    to_wintertime : bool, optional
        if True the time index is converted to Dutch winter time. The default
        is True.
    drop_duplicate_times : bool, optional
        if True rows with a duplicate time stamp are removed keeping only the
        first row. The default is True.
    add_registration_history : bool, optional
        if True the registration history is added to the metadata. The defualt
        is True.

    Raises
    ------
    ValueError
        if bro_id is invalid.

    Returns
    -------
    df : pd.DataFrame
        measurements.
    meta : dict
        metadata.

    """
    if not bro_id.startswith("GLD"):
        raise ValueError("can only get observations if bro id starts with GLD")

    url = "https://publiek.broservices.nl/gm/gld/v1/objects/{}"
    params = {}
    if tmin is not None:
        tmin = pd.to_datetime(tmin)
        params["observationPeriodBeginDate"] = tmin.strftime("%Y-%m-%d")
    if tmax is not None:
        tmax = pd.to_datetime(tmax)
        params["observationPeriodEndDate"] = tmax.strftime("%Y-%m-%d")

    # add some logic to retry in case of a 429 response
    s = requests.Session()
    retries = Retry(total=5, backoff_factor=0.5, status_forcelist=[429])
    s.mount("https://", HTTPAdapter(max_retries=retries))
    req = s.get(url.format(bro_id), params=params)

    if req.status_code > 200:
        req.raise_for_status()

    ns = {
        "ns11": "http://www.broservices.nl/xsd/dsgld/1.0",
        "gldcommon": "http://www.broservices.nl/xsd/gldcommon/1.0",
        "waterml": "http://www.opengis.net/waterml/2.0",
        "swe": "http://www.opengis.net/swe/2.0",
        "om": "http://www.opengis.net/om/2.0",
    }

    tree = xml.etree.ElementTree.fromstring(req.text)

    glds = tree.findall(".//ns11:GLD_O", ns)
    if len(glds) != 1:
        raise (Exception("Only one gld supported"))
    gld = glds[0]

    meta = {"source": "BRO"}
    meta["location"] = gld.find("ns11:monitoringPoint//gldcommon:broId", ns).text
    meta["tube_nr"] = int(
        gld.find("ns11:monitoringPoint//gldcommon:tubeNumber", ns).text
    )
    meta["name"] = f"{meta['location']}_{meta['tube_nr']}"
    gmn = gld.find("ns11:groundwaterMonitoringNet//gldcommon:broId", ns)
    if gmn is None:
        meta["monitoringsnet"] = None
    else:
        meta["monitoringsnet"] = gmn.text

    # get observations
    msts = "ns11:observation//om:result//waterml:MeasurementTimeseries"
    times = [time.text for time in gld.findall(f"{msts}//waterml:time", ns)]
    values = [
        np.nan if value.text is None else float(value.text)
        for value in gld.findall(f"{msts}//waterml:value", ns)
    ]
    qualifiers = [q.text for q in gld.findall(f"{msts}//swe:Category//swe:value", ns)]

    # to dataframe
    df = pd.DataFrame(
        index=pd.to_datetime(times, utc=True).tz_convert("CET"),
        data={"values": values, "qualifier": qualifiers},
    )

    # wintertime
    if to_wintertime:
        # remove time zone information by transforming to dutch winter time
        df.index = pd.to_datetime(df.index, utc=True).tz_localize(None) + pd.Timedelta(
            1, unit="h"
        )

    # duplicates
    if df.index.has_duplicates and drop_duplicate_times:
        duplicates = df.index.duplicated(keep="first")
        message = "{} contains {} duplicates (of {}). Keeping only first values."
        logger.debug(message.format(bro_id, duplicates.sum(), len(df)))
        df = df[~duplicates]

    df = df.sort_index()

    # slice to tmin and tmax
    df = df.loc[tmin:tmax]

    # add metadata from gmw
    meta.update(get_metadata_from_gmw(meta["location"], meta["tube_nr"]))

    return df, meta


def get_full_metadata_from_gmw(bro_id, tube_nr):
    """get metadata for a groundwater monitoring well.


    Parameters
    ----------
    bro_id : str
        bro id of groundwater monitoring well e.g. 'GMW000000036287'.
    tube_nr : int
        filter number you want metadata for.

    Raises
    ------
    ValueError
        if bro_id is invalid.

    Returns
    -------
    meta : dict
        dictionary with metadata.

    """

    if not bro_id.startswith("GMW"):
        raise ValueError("can only get metadata if bro id starts with GMW")

    url = f"https://publiek.broservices.nl/gm/gmw/v1/objects/{bro_id}"
    req = requests.get(url)

    # read results
    tree = xml.etree.ElementTree.fromstring(req.text)
    ns = "{http://www.broservices.nl/xsd/dsgmw/1.1}"

    gmws = tree.findall(f".//{ns}GMW_PO")
    if len(gmws) != 1:
        raise (Exception("Only one gmw supported"))
    gmw = gmws[0]
    meta = {"location": bro_id, "tube_nr": tube_nr, "source": "BRO"}
    for child in gmw:
        key = child.tag.split("}", 1)[1]
        if len(child) == 0:
            meta[key] = child.text
        elif key == "deliveredLocation":
            ns = "{http://www.broservices.nl/xsd/gmwcommon/1.1}"
            point = child.find(f"{ns}location")
            ns = "{http://www.opengis.net/gml/3.2}"
            xy = [float(x) for x in point.find(f"{ns}pos").text.split()]
            meta["x"], meta["y"] = xy
        elif key in ["wellConstructionDate"]:
            meta[key] = child[0].text
        elif key == "wellHistory":
            for grandchild in child:
                key = grandchild.tag.split("}", 1)[1]
                meta[key] = grandchild[0].text
        elif key in ["deliveredVerticalPosition", "registrationHistory"]:
            for grandchild in child:
                key = grandchild.tag.split("}", 1)[1]
                meta[key] = grandchild.text
        elif key in ["monitoringTube"]:
            tube = False
            tube_dic = {}
            for grandchild in child:
                if len(grandchild) == 0:
                    tube_key = grandchild.tag.split("}", 1)[1]
                    if tube_key == "tubeNumber":
                        if int(grandchild.text) == tube_nr:
                            tube = True
                    else:
                        tube_dic[tube_key] = grandchild.text

                else:
                    for greatgrandchild in grandchild:
                        tube_key = greatgrandchild.tag.split("}", 1)[1]
                        tube_dic[tube_key] = greatgrandchild.text
            if tube:
                meta.update(tube_dic)

    rename_dic = {
        "groundLevelPosition": "ground_level",
        "screenTopPosition": "screen_top",
        "screenBottomPosition": "screen_bottom",
        "tubeTopPosition": "tube_top",
    }

    for key, val in rename_dic.items():
        meta[val] = meta.pop(key)

    return meta


@lru_cache()
def _get_gmw_from_bro_id(bro_id, retries=0):
    """get a gmw object from a bro_id

    Parameters
    ----------
    bro_id : str
        bro id of groundwater monitoring well e.g. 'GMW000000036287'.

    Returns
    -------
    Element
        xml reference to gmw object

    Raises
    ------
    ValueError
        if bro_id is invalid.
    """

    if not bro_id.startswith("GMW"):
        raise ValueError("can only get metadata if bro id starts with GMW")

    ns = {
        "dsgmw": "http://www.broservices.nl/xsd/dsgmw/1.1",
        "gmwcommon": "http://www.broservices.nl/xsd/gmwcommon/1.1",
        "gml": "http://www.opengis.net/gml/3.2",
    }

    url = f"https://publiek.broservices.nl/gm/gmw/v1/objects/{bro_id}"
    req = requests.get(url)

    # read results
    tree = xml.etree.ElementTree.fromstring(req.text)

    gmws = tree.findall(".//dsgmw:GMW_PO", ns)
    if len(gmws) != 1:
        max_retries = rcParams["bro"]["max_retries"]
        val_ind = req.text.find("valid")
        valid = req.text[(val_ind + 9) : (val_ind + 14)]
        if valid == "false" and retries < max_retries:
            logger.debug(
                f"got invalid response for {bro_id}, trying again {retries + 1}/{max_retries}"
            )
            return _get_gmw_from_bro_id(bro_id, retries=retries + 1)
        elif valid == "false":
            raise Exception(
                f"got invalid response for {bro_id} after trying {retries} times"
            )
        raise (Exception("Only one gmw supported"))
    gmw = gmws[0]

    return gmw


def get_tube_nrs_from_gmw(bro_id):
    """returns all tube numbers from a groundwater monitoring well (gmw)

    Parameters
    ----------
    bro_id : str
        bro id of groundwater monitoring well e.g. 'GMW000000036287'.

    Returns
    -------
    list of int
        tube numbers
    """
    ns = {
        "dsgmw": "http://www.broservices.nl/xsd/dsgmw/1.1",
        "gmwcommon": "http://www.broservices.nl/xsd/gmwcommon/1.1",
        "gml": "http://www.opengis.net/gml/3.2",
    }

    gmw = _get_gmw_from_bro_id(bro_id)

    # get tube nrs
    tube_numbers = [
        int(tube.text)
        for tube in gmw.findall("dsgmw:monitoringTube/dsgmw:tubeNumber", ns)
    ]

    return tube_numbers


def get_metadata_from_gmw(bro_id, tube_nr):
    """get selection of metadata for a groundwater monitoring well.
    coordinates, ground_level, tube_top and tube screen

    Parameters
    ----------
    bro_id : str
        bro id of groundwater monitoring well e.g. 'GMW000000036287'.
    tube_nr : int
        tube number you want metadata for.

    Raises
    ------
    ValueError
        if bro_id is invalid.
    TypeError
        if tube_nr is not an int

    Returns
    -------
    meta : dict
        dictionary with metadata.

    """
    if not isinstance(tube_nr, int):
        raise TypeError(f"expected integer got {type(tube_nr)}")

    ns = {
        "dsgmw": "http://www.broservices.nl/xsd/dsgmw/1.1",
        "gmwcommon": "http://www.broservices.nl/xsd/gmwcommon/1.1",
        "gml": "http://www.opengis.net/gml/3.2",
    }

    gmw = _get_gmw_from_bro_id(bro_id)

    meta = {"location": bro_id, "tube_nr": tube_nr, "source": "BRO"}

    # x and y
    xy_elem = gmw.find("dsgmw:deliveredLocation//gmwcommon:location//gml:pos", ns)
    xy = [float(val) for val in xy_elem.text.split()]

    # convert crs
    srsname = gmw.find("dsgmw:deliveredLocation//gmwcommon:location", ns).attrib[
        "srsName"
    ]
    epsg_gwm = int(srsname.split(":")[-1])
    proj_from = Proj(f"EPSG:{epsg_gwm}")
    proj_to = Proj(EPSG_28992)
    transformer = Transformer.from_proj(proj_from, proj_to)
    xy = transformer.transform(xy[0], xy[1])

    meta["x"], meta["y"] = xy

    # ground_level
    vert_pos = gmw.find("dsgmw:deliveredVerticalPosition", ns)
    mv = vert_pos.find("gmwcommon:groundLevelPosition", ns)
    datum = vert_pos.find("gmwcommon:verticalDatum", ns)
    if datum.text == "NAP" and mv.attrib["uom"] == "m":
        meta["unit"] = "m NAP"
        if mv.text is None:
            meta["ground_level"] = np.nan
        else:
            meta["ground_level"] = float(mv.text)
    else:
        raise ValueError("invalid ground_level datum or unit")

    # buis eigenschappen
    tubes = gmw.findall("dsgmw:monitoringTube", ns)
    tube_nrs = [int(tube.find("dsgmw:tubeNumber", ns).text) for tube in tubes]
    if tube_nr not in tube_nrs:
        raise ValueError(
            f"gmw {bro_id} has no tube_nr {tube_nr} please choose a tube_nr from"
            f"{tube_nrs}"
        )
    tube = tubes[tube_nrs.index(tube_nr)]

    # tube_top
    mp = tube.find("dsgmw:tubeTopPosition", ns)
    if mp.attrib["uom"] == "m":
        meta["tube_top"] = float(mp.text)

    # bovenkant filter
    bkf = tube.find("dsgmw:screen//dsgmw:screenTopPosition", ns)
    if bkf.attrib["uom"] == "m":
        meta["screen_top"] = float(bkf.text)

    # onderkant filter
    okf = tube.find("dsgmw:screen//dsgmw:screenBottomPosition", ns)
    if okf.attrib["uom"] == "m":
        meta["screen_bottom"] = float(okf.text)

    meta["metadata_available"] = True

    return meta


def get_obs_list_from_extent(
    extent,
    ObsClass,
    tmin=None,
    tmax=None,
    only_metadata=False,
    keep_all_obs=True,
    epsg=28992,
    ignore_max_obs=False,
):
    """get a list of gmw observations within an extent.


    Parameters
    ----------
    extent : list, tuple, numpy-array or None, optional
        get groundwater monitoring wells within this extent
        [xmin, xmax, ymin, ymax]
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    tmin : str or None, optional
        start time of observations. The default is None.
    tmax : str or None, optional
        end time of observations. The default is None.
    only_metadata : bool, optional
        if True download only metadata, significantly faster. The default
        is False.
    epsg : int, optional
        epsg code of the extent. The default is 28992 (RD).
    ignore_max_obs : bool, optional
        by default you get a prompt if you want to download over a 1000
        observations at once. if ignore_max_obs is True you won't get the
        prompt. The default is False

    Raises
    ------

        DESCRIPTION.

    Returns
    -------
    obs_list : TYPE
        DESCRIPTION.

    """

    if only_metadata and not keep_all_obs:
        logger.error(
            "you will get an empty ObsCollection with only_metadata is True and"
            "keep_all_obs is False"
        )

    url = "https://publiek.broservices.nl/gm/gmw/v1/characteristics/searches?"

    data = {}
    if tmin is None or tmax is None:
        data["registrationPeriod"] = {}
        if tmin is not None:
            beginDate = pd.to_datetime(tmin).strftime("%Y-%m-%d")
            data["registrationPeriod"]["beginDate"] = beginDate
        if tmax is not None:
            endDate = pd.to_datetime(tmax).strftime("%Y-%m-%d")
            data["registrationPeriod"]["endDate"] = endDate

    data["area"] = {}
    if epsg == 4326:
        data["area"]["boundingBox"] = {
            "lowerCorner": {"lat": extent[2], "lon": extent[0]},
            "upperCorner": {"lat": extent[3], "lon": extent[1]},
        }
    else:
        transformer = Transformer.from_crs(epsg, 4326)
        if extent is not None:
            lat1, lon1 = transformer.transform(extent[0], extent[2])
            lat2, lon2 = transformer.transform(extent[1], extent[3])
            data["area"]["boundingBox"] = {
                "lowerCorner": {"lat": lat1, "lon": lon1},
                "upperCorner": {"lat": lat2, "lon": lon2},
            }
    req = requests.post(url, json=data)
    if req.status_code > 200:
        logger.error(
            "could not get monitoring wells, your extent is probably too big."
            "Try a smaller extent"
        )
        req.raise_for_status()

    # read results
    tree = xml.etree.ElementTree.fromstring(req.text)

    ns = {
        "dsgmw": "http://www.broservices.nl/xsd/dsgmw/1.1",
        "gml": "http://www.opengis.net/gml/3.2",
        "brocom": "http://www.broservices.nl/xsd/brocommon/3.0",
    }

    if tree.find(".//brocom:responseType", ns).text == "rejection":
        raise RuntimeError(tree.find(".//brocom:rejectionReason", ns).text)

    gmws_ids = np.unique(
        [gmw.text for gmw in tree.findall(".//dsgmw:GMW_C//brocom:broId", ns)]
    )

    if len(gmws_ids) > 1000 and not ignore_max_obs:
        ans = input(
            f"You requested to download {len(gmws_ids)} observations, this can"
            "take a while. Are you sure you want to continue [Y/n]? "
        )
        if ans not in ["Y", "y", "yes", "Yes", "YES"]:
            return []

    obs_list = []
    for gmw_id in tqdm(gmws_ids):
        gmws = tree.findall(f'.//*[brocom:broId="{gmw_id}"]', ns)
        if len(gmws) < 1:
            raise RuntimeError("unexpected")

        tube_nrs = get_tube_nrs_from_gmw(gmw_id)
        for tube_nr in tube_nrs:
            o = ObsClass.from_bro(
                gmw_id,
                tube_nr=tube_nr,
                tmin=tmin,
                tmax=tmax,
                only_metadata=only_metadata,
            )
            if o.empty:
                logger.debug(
                    f"no measurements found for gmw_id {gmw_id} and tube number"
                    f"{tube_nr}"
                )
                if keep_all_obs:
                    obs_list.append(o)
            else:
                obs_list.append(o)

    return obs_list
